from typing import Optional
from pathlib import Path
import sqlalchemy as sql
from sqlalchemy_utils.functions import database_exists, create_database, drop_database


class Env:
    """
    Store for runtime globals.
    """
    _instance: Optional['Env'] = None

    @classmethod
    def get(cls) -> 'Env':
        if cls._instance is None:
            cls._instance = Env()
        return cls._instance

    def __init__(self):
        self._home: Optional[Path] = None
        self._db_path: Optional[Path] = None
        self._img_dir: Optional[Path] = None
        self._nnidx_dir: Optional[Path] = None
        self._tmp_video_dir: Optional[Path] = None
        self._sa_engine: Optional[sql.engine.base.Engine] = None
        self._db_name: Optional[str] = None

    def set_up(self, home_str: Optional[str], db_name: Optional[str], echo: bool = False) -> None:
        home = Path.home() / '.pixeltable' if home_str is None else Path(home_str)
        if db_name is None:
            db_name = 'pixeltable'
        self.set_home(home)
        if self._home.exists() and not self._home.is_dir():
            raise RuntimeError(f'{self._home} is not a directory')

        self._db_name = db_name
        db_url = f'postgresql:///{self._db_name}'

        if not self._home.exists():
            print(f'setting up Pixeltable at {self._home}, db at {db_url}')
            self._home.mkdir()
            self._img_dir.mkdir()
            self._nnidx_dir.mkdir()
            self._tmp_video_dir.mkdir()
            self.tear_down()
            if not database_exists(db_url):
                create_database(db_url)
            self._sa_engine = sql.create_engine(db_url, echo=echo, future=True)
            from pixeltable import store
            store.Base.metadata.create_all(self._sa_engine)
        else:
            if not database_exists(db_url):
                raise RuntimeError(f'Database not found: {db_url}')
            if self._sa_engine is None:
                self._sa_engine = sql.create_engine(db_url, echo=echo, future=True)

    def tear_down(self) -> None:
        db_url = f'postgresql:///{self._db_name}'
        if database_exists(db_url):
            drop_database(db_url)

    def set_home(self, home: Path) -> None:
        if self._home is not None:
            return
        self._home = home
        self._db_path = self._home / 'db.sqlite3'
        self._img_dir = self._home / 'images'
        self._nnidx_dir = self._home / 'nnidxs'
        self._tmp_video_dir = self._home / 'tmp_videos'

    @property
    def img_dir(self) -> Path:
        assert self._img_dir is not None
        return self._img_dir

    @property
    def nnidx_dir(self) -> Path:
        assert self._nnidx_dir is not None
        return self._nnidx_dir

    @property
    def tmp_video_dir(self) -> Path:
        assert self._tmp_video_dir is not None
        return self._tmp_video_dir

    @property
    def engine(self) -> sql.engine.base.Engine:
        assert self._sa_engine is not None
        return self._sa_engine



import pytest
import math

import pixeltable as pt
from pixeltable import exceptions as exc
from pixeltable import catalog
from pixeltable.type_system import \
    StringType, IntType, FloatType, TimestampType, ImageType, VideoType, JsonType, BoolType
from pixeltable.tests.utils import make_tbl, create_table_data, read_data_file, get_video_files, sum_uda
from pixeltable.functions import make_video
from pixeltable import utils


class TestTable:
    def test_create(self, test_db: catalog.Db) -> None:
        db = test_db
        db.create_dir('dir1')
        c1 = catalog.Column('c1', StringType(), nullable=False)
        c2 = catalog.Column('c2', IntType(), nullable=False)
        c3 = catalog.Column('c3', FloatType(), nullable=False)
        c4 = catalog.Column('c4', TimestampType(), nullable=False)
        schema = [c1, c2, c3, c4]
        _ = db.create_table('test', schema)
        _ = db.create_table('dir1.test', schema)

        with pytest.raises(exc.BadFormatError):
            _ = db.create_table('1test', schema)
        with pytest.raises(exc.BadFormatError):
            _ = catalog.Column('1c', StringType())
        with pytest.raises(exc.DuplicateNameError):
            _ = db.create_table('test2', [c1, c1])
        with pytest.raises(exc.DuplicateNameError):
            _ = db.create_table('test', schema)
        with pytest.raises(exc.DuplicateNameError):
            _ = db.create_table('test2', [c1, c1])
        with pytest.raises(exc.UnknownEntityError):
            _ = db.create_table('dir2.test2', schema)

        _ = db.list_tables()
        _ = db.list_tables('dir1')

        with pytest.raises(exc.BadFormatError):
            _ = db.list_tables('1dir')
        with pytest.raises(exc.UnknownEntityError):
            _ = db.list_tables('dir2')

        # test loading with new client
        cl2 = pt.Client()
        db = cl2.get_db('test')

        tbl = db.get_table('test')
        assert isinstance(tbl, catalog.MutableTable)
        tbl.add_column(catalog.Column('c5', IntType()))
        tbl.drop_column('c1')
        tbl.rename_column('c2', 'c17')

        db.rename_table('test', 'test2')

        db.drop_table('test2')
        db.drop_table('dir1.test')

        with pytest.raises(exc.UnknownEntityError):
            db.drop_table('test')
        with pytest.raises(exc.UnknownEntityError):
            db.drop_table('dir1.test2')
        with pytest.raises(exc.BadFormatError):
            db.drop_table('.test2')

    def test_create_images(self, test_db: catalog.Db) -> None:
        db = test_db
        cols = [
            catalog.Column('img', ImageType(), nullable=False),
            catalog.Column('category', StringType(), nullable=False),
            catalog.Column('split', StringType(), nullable=False),
        ]
        tbl = db.create_table('test', cols)
        df = read_data_file('imagenette2-160', 'manifest.csv', ['img'])
        # TODO: insert a random subset
        tbl.insert_pandas(df[:20])
        html_str = tbl.show(n=100)._repr_html_()
        print(html_str)
        # TODO: check html_str

    def test_create_video(self, test_db: catalog.Db) -> None:
        db = test_db
        cols = [
            catalog.Column('video', VideoType(), nullable=False),
            catalog.Column('frame', ImageType(), nullable=False),
            catalog.Column('frame_idx', IntType(), nullable=False),
        ]
        tbl = db.create_table(
            'test', cols, extract_frames_from='video', extracted_frame_col='frame',
            extracted_frame_idx_col='frame_idx', extracted_fps=0)
        params = tbl.parameters
        # reload to make sure that metadata gets restored correctly
        cl = pt.Client()
        db = cl.get_db('test')
        tbl = db.get_table('test')
        assert tbl.parameters == params
        tbl.insert_rows([[get_video_files()[0]]], ['video'])
        html_str = tbl.show(n=100)._repr_html_()
        # TODO: check html_str
        _ = tbl[make_video(tbl.frame_idx, tbl.frame)].group_by(tbl.video).show()

        with pytest.raises(exc.Error):
            # can't drop frame col
            tbl.drop_column('frame')
        with pytest.raises(exc.Error):
            # can't drop frame_idx col
            tbl.drop_column('frame_idx')
        with pytest.raises(exc.BadFormatError):
            # missing parameters
            _ = db.create_table(
                'exc', cols, extract_frames_from='video',
                extracted_frame_idx_col='frame_idx', extracted_fps=0)
        with pytest.raises(exc.BadFormatError):
            # wrong column type
            _ = db.create_table(
                'exc', cols, extract_frames_from='frame', extracted_frame_col='frame',
                extracted_frame_idx_col='frame_idx', extracted_fps=0)
        with pytest.raises(exc.BadFormatError):
            # wrong column type
            _ = db.create_table(
                'exc', cols, extract_frames_from='video', extracted_frame_col='frame_idx',
                extracted_frame_idx_col='frame_idx', extracted_fps=0)
        with pytest.raises(exc.BadFormatError):
            # wrong column type
            _ = db.create_table(
                'exc', cols, extract_frames_from='video', extracted_frame_col='frame',
                extracted_frame_idx_col='frame', extracted_fps=0)
        with pytest.raises(exc.BadFormatError):
            # unknown column
            _ = db.create_table(
                'exc', cols, extract_frames_from='breaks', extracted_frame_col='frame',
                extracted_frame_idx_col='frame_idx', extracted_fps=0)
        with pytest.raises(exc.BadFormatError):
            # unknown column
            _ = db.create_table(
                'exc', cols, extract_frames_from='video', extracted_frame_col='breaks',
                extracted_frame_idx_col='frame_idx', extracted_fps=0)
        with pytest.raises(exc.BadFormatError):
            # unknown column
            _ = db.create_table(
                'exc', cols, extract_frames_from='video', extracted_frame_col='frame',
                extracted_frame_idx_col='breaks', extracted_fps=0)

    def test_insert(self, test_db: catalog.Db) -> None:
        db = test_db
        t1 = make_tbl(db, 'test1', ['c1', 'c2'])
        data1 = create_table_data(t1)
        t1.insert_pandas(data1)
        assert t1.count() == len(data1)

        # incompatible schema
        t2 = make_tbl(db, 'test2', ['c2', 'c1'])
        t2_data = create_table_data(t2)
        with pytest.raises(exc.InsertError):
            t1.insert_pandas(t2_data)

    def test_query(self, test_db: catalog.Db) -> None:
        db = test_db
        t = make_tbl(db, 'test', ['c1', 'c2', 'c3', 'c4', 'c5'])
        t_data = create_table_data(t)
        t.insert_pandas(t_data)
        _ = t.show(n=0)

        # test querying existing table
        cl2 = pt.Client()
        db2 = cl2.get_db('test')
        t2 = db2.get_table('test')
        _  = t2.show(n=0)

    def test_computed_cols(self, test_db: catalog.Db) -> None:
        db = test_db
        c1 = catalog.Column('c1', IntType(), nullable=False)
        c2 = catalog.Column('c2', FloatType(), nullable=False)
        c3 = catalog.Column('c3', JsonType(), nullable=False)
        schema = [c1, c2, c3]
        t = db.create_table('test', schema)
        t.add_column(catalog.Column('c4', computed_with=t.c1 + 1))
        t.add_column(catalog.Column('c5', computed_with=t.c4 + 1))
        t.add_column(catalog.Column('c6', computed_with=t.c1 / t.c2))
        t.add_column(catalog.Column('c7', computed_with=t.c6 * t.c2))
        t.add_column(catalog.Column('c8', computed_with=t.c3.detections['*'].bounding_box))
        t.add_column(catalog.Column('c9', FloatType(), computed_with=lambda c2: math.sqrt(c2)))

        # Column.dependent_cols are computed correctly
        assert len(t.c1.col.dependent_cols) == 2
        assert len(t.c2.col.dependent_cols) == 3
        assert len(t.c3.col.dependent_cols) == 1
        assert len(t.c4.col.dependent_cols) == 1
        assert len(t.c5.col.dependent_cols) == 0
        assert len(t.c6.col.dependent_cols) == 1
        assert len(t.c7.col.dependent_cols) == 0
        assert len(t.c8.col.dependent_cols) == 0

        data_df = create_table_data(t, ['c1', 'c2', 'c3'], num_rows=10)
        t.insert_pandas(data_df)
        _ = t.show()

        # not allowed to pass values for computed cols
        with pytest.raises(exc.InsertError):
            data_df2 = create_table_data(t, num_rows=10)
            t.insert_pandas(data_df2)

        # computed col references non-existent col
        with pytest.raises(exc.Error):
            c1 = catalog.Column('c1', IntType(), nullable=False)
            c2 = catalog.Column('c2', FloatType(), nullable=False)
            c3 = catalog.Column('c3', FloatType(), nullable=False, computed_with=lambda c2: math.sqrt(c2))
            _ = db.create_table('test2', [c1, c3, c2])

        # test loading from store
        cl2 = pt.Client()
        db2 = cl2.get_db('test')
        t2 = db2.get_table('test')
        assert len(t.columns) == len(t2.columns)
        for i in range(len(t.columns)):
            if t.columns[i].value_expr is not None:
                assert t.columns[i].value_expr.equals(t2.columns[i].value_expr)

        # make sure we can still insert data and that computed cols are still set correctly
        t2.insert_pandas(data_df)
        res = t2.show(0)
        tbl_df = t2.show(0).to_pandas()

        # can't drop c4: c5 depends on it
        with pytest.raises(exc.Error):
            t.drop_column('c4')
        t.drop_column('c5')
        # now it works
        t.drop_column('c4')

    def test_computed_img_cols(self, test_db: catalog.Db) -> None:
        db = test_db
        c1 = catalog.Column('img', ImageType(), nullable=False, indexed=True)
        schema = [c1]
        t = db.create_table('test', schema)
        t.add_column(catalog.Column('c2', computed_with=t.img.width))
        t.add_column(catalog.Column('c3', computed_with=t.img.rotate(90)))

        data_df = read_data_file('imagenette2-160', 'manifest.csv', ['img'])
        t.insert_pandas(data_df.loc[0:20, ['img']])
        _ = t.show()
        assert utils.computed_img_count(tbl_id=t.id) == t.count()

        # test loading from store
        cl2 = pt.Client()
        db2 = cl2.get_db('test')
        t2 = db2.get_table('test')
        assert len(t.columns) == len(t2.columns)
        for i in range(len(t.columns)):
            if t.columns[i].value_expr is not None:
                assert t.columns[i].value_expr.equals(t2.columns[i].value_expr)

        # make sure we can still insert data and that computed cols are still set correctly
        t2.insert_pandas(data_df.loc[0:20, ['img']])
        assert utils.computed_img_count(tbl_id=t.id) == t2.count()
        res = t2.show(0)
        tbl_df = t2.show(0).to_pandas()
        print(tbl_df)

        # revert also removes computed images
        t2.revert()
        assert utils.computed_img_count() == t2.count()

    def test_computed_window_fn(self, test_db: catalog.Db, test_tbl: catalog.Table) -> None:
        db = test_db
        t = test_tbl
        # backfill
        t.add_column(catalog.Column('c9', computed_with=sum_uda(t.c2).window(partition_by=t.c4, order_by=t.c3)))

        c2 = catalog.Column('c2', IntType(), nullable=False)
        c3 = catalog.Column('c3', FloatType(), nullable=False)
        c4 = catalog.Column('c4', BoolType(), nullable=False)
        new_t = db.create_table('insert_test', [c2, c3, c4])
        new_t.add_column(catalog.Column('c5', IntType(), computed_with=lambda c2: c2 * c2))
        new_t.add_column(catalog.Column(
            'c6', computed_with=sum_uda(new_t.c5).window(partition_by=new_t.c4, order_by=new_t.c3)))
        data_df = t[t.c2, t.c4, t.c3].show(0).to_pandas()
        new_t.insert_pandas(data_df)
        _ = new_t.show(0)
        print(_)

    def test_revert(self, test_db: catalog.Db) -> None:
        db = test_db
        t1 = make_tbl(db, 'test1', ['c1', 'c2'])
        data1 = create_table_data(t1)
        t1.insert_pandas(data1)
        assert t1.count() == len(data1)
        data2 = create_table_data(t1)
        t1.insert_pandas(data2)
        assert t1.count() == len(data1) + len(data2)
        t1.revert()
        assert t1.count() == len(data1)
        t1.insert_pandas(data2)
        assert t1.count() == len(data1) + len(data2)

    def test_snapshot(self, test_db: catalog.Db) -> None:
        db = test_db
        db.create_dir('main')
        tbl = make_tbl(db, 'main.test1', ['c1', 'c2'])
        data1 = create_table_data(tbl)
        tbl.insert_pandas(data1)
        assert tbl.count() == len(data1)

        db.create_snapshot('snap', ['main.test1'])
        snap = db.get_table('snap.test1')
        assert snap.count() == len(data1)

        # adding data to a base table doesn't change the snapshot
        data2 = create_table_data(tbl)
        tbl.insert_pandas(data2)
        assert tbl.count() == len(data1) + len(data2)
        assert snap.count() == len(data1)

        tbl.revert()
        # can't revert a version referenced by a snapshot
        with pytest.raises(exc.OperationalError):
            tbl.revert()

    def test_add_column(self, test_db: catalog.Db) -> None:
        db = test_db
        t = make_tbl(db, 'test', ['c1', 'c2'])
        data1 = create_table_data(t)
        t.insert_pandas(data1)
        assert t.count() == len(data1)
        t.add_column(catalog.Column('c3', computed_with=t.c2 + 10, nullable=False))
        _ = t.show()
        print(_)



from typing import List, Dict
import os

import sqlalchemy.orm as orm

from pixeltable import catalog, store
from pixeltable.env import Env

__all__ = [
    'Client',
]


class Client:
    """Missing docstring.
    """

    def __init__(self) -> None:
        self.db_cache: Dict[str, catalog.Db] = {}
        Env.get().set_up(os.environ.get('PIXELTABLE_HOME'), os.environ.get('PIXELTABLE_DB'))

    def create_db(self, name: str) -> catalog.Db:
        db = catalog.Db.create(name)
        self.db_cache[name] = db
        return db

    def drop_db(self, name: str, force: bool = False) -> None:
        if not force:
            return
        if name in self.db_cache:
            self.db_cache[name].delete()
            del self.db_cache[name]
        else:
            catalog.Db.load(name).delete()

    def get_db(self, name: str) -> catalog.Db:
        if name in self.db_cache:
            return self.db_cache[name]
        db = catalog.Db.load(name)
        self.db_cache[name] = db
        return db

    def list_dbs(self) -> List[str]:
        with orm.Session(store.engine) as session:
            return [r[0] for r in session.query(store.Db.name)]



import sqlalchemy as sql
import pytest

from pixeltable import catalog
from pixeltable.type_system import StringType, BoolType, IntType, ImageType, ArrayType, ColumnType, FloatType
from pixeltable.function import Function
from pixeltable.exprs import Expr, CompoundPredicate, FunctionCall, Literal, InlineDict, InlineArray, ColumnRef
from pixeltable.exprs import RELATIVE_PATH_ROOT as R
from pixeltable.functions import udf_call, dict_map, cast, sum, count
from pixeltable.functions.pil.image import blend
from pixeltable.functions.clip import encode_image
from pixeltable import exceptions as exc
from pixeltable.tests import utils


class TestExprs:
    # This breaks with exception 'cannot pickle _thread._local obj'
    # sum = Function(
    #     IntType(), [IntType()],
    #     init_fn=lambda: TestExprs.SumAggregator(), update_fn=SumAggregator.update, value_fn=SumAggregator.value)

    def test_basic(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        assert isinstance(t['c1'] < 'a', Expr)
        assert isinstance(t.c1 < 'a', Expr)
        assert isinstance(t['c1'] <= 'a', Expr)
        assert isinstance(t.c1 <= 'a', Expr)
        assert isinstance(t['c1'] == 'a', Expr)
        assert isinstance(t.c1 == 'a', Expr)
        assert isinstance(t['c1'] != 'a', Expr)
        assert isinstance(t.c1 != 'a', Expr)
        assert isinstance(t['c1'] > 'a', Expr)
        assert isinstance(t.c1 > 'a', Expr)
        assert isinstance(t['c1'] >= 'a', Expr)
        assert isinstance(t.c1 >= 'a', Expr)
        assert isinstance((t.c1 == 'a') & (t.c2 < 5), Expr)
        assert isinstance((t.c1 == 'a') | (t.c2 < 5), Expr)
        assert isinstance(~(t.c1 == 'a'), Expr)

    def test_compound_predicates(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        # compound predicates that can be fully evaluated in SQL
        e = ((t.c1 == 'test string') & (t.c2 > 50)).sql_expr()
        assert len(e.clauses) == 2

        e = ((t.c1 == 'test string') & (t.c2 > 50) & (t.c3 < 1.0)).sql_expr()
        assert len(e.clauses) == 3
        e = ((t.c1 == 'test string') | (t.c2 > 50)).sql_expr()
        assert len(e.clauses) == 2
        e = ((t.c1 == 'test string') | (t.c2 > 50) | (t.c3 < 1.0)).sql_expr()
        assert len(e.clauses) == 3
        e = (~(t.c1 == 'test string')).sql_expr()
        assert isinstance(e, sql.sql.expression.BinaryExpression)

        # compound predicates with Python functions
        udf = Function(BoolType(), [StringType()], eval_fn=lambda a: True)
        udf2 = Function(BoolType(), [IntType()], eval_fn=lambda a: True)

        # & can be split
        p = (t.c1 == 'test string') & udf(t.c1)
        assert p.sql_expr() is None
        sql_pred, other_pred = p.extract_sql_predicate()
        assert isinstance(sql_pred, sql.sql.expression.BinaryExpression)
        assert isinstance(other_pred, FunctionCall)

        p = (t.c1 == 'test string') & udf(t.c1) & (t.c2 > 50)
        assert p.sql_expr() is None
        sql_pred, other_pred = p.extract_sql_predicate()
        assert len(sql_pred.clauses) == 2
        assert isinstance(other_pred, FunctionCall)

        p = (t.c1 == 'test string') & udf(t.c1) & (t.c2 > 50) & udf2(t.c2)
        assert p.sql_expr() is None
        sql_pred, other_pred = p.extract_sql_predicate()
        assert len(sql_pred.clauses) == 2
        assert isinstance(other_pred, CompoundPredicate)

        # | cannot be split
        p = (t.c1 == 'test string') | udf(t.c1)
        assert p.sql_expr() is None
        sql_pred, other_pred = p.extract_sql_predicate()
        assert sql_pred is None
        assert isinstance(other_pred, CompoundPredicate)

    def test_basic_filter(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        _ = t[t.c1 == 'test string'].show()
        print(_)
        _ = t[t.c2 > 50].show()
        print(_)

    def test_arithmetic_exprs(self, test_tbl: catalog.Table) -> None:
        t = test_tbl

        for op1, op2 in [(t.c2, t.c2), (t.c3, t.c3)]:
            _ = t[op1 + op2].show()
            _ = t[op1 - op2].show()
            _ = t[op1 * op2].show()
            _ = t[op1 > 0][op1 / op2].show()

        # non-numeric types
        for op1, op2 in [
            (t.c1, t.c2), (t.c1, 1), (t.c2, t.c1), (t.c2, 'a'),
            (t.c1, t.c3), (t.c1, 1.0), (t.c3, t.c1), (t.c3, 'a')
        ]:
            with pytest.raises(exc.OperationalError):
                _ = t[op1 + op2]
            with pytest.raises(exc.OperationalError):
                _ = t[op1 - op2]
            with pytest.raises(exc.OperationalError):
                _ = t[op1 * op2]
            with pytest.raises(exc.OperationalError):
                _ = t[op1 / op2]

        # TODO: test division; requires predicate
        for op1, op2 in [(t.c6.f2, t.c6.f2), (t.c6.f3, t.c6.f3)]:
            _ = t[op1 + op2].show()
            _ = t[op1 - op2].show()
            _ = t[op1 * op2].show()

        for op1, op2 in [
            (t.c6.f1, t.c6.f2), (t.c6.f1, t.c6.f3), (t.c6.f1, 1), (t.c6.f1, 1.0),
            (t.c6.f2, t.c6.f1), (t.c6.f3, t.c6.f1), (t.c6.f2, 'a'), (t.c6.f3, 'a'),
        ]:
            with pytest.raises(exc.OperationalError):
                _ = t[op1 + op2].show()
            with pytest.raises(exc.OperationalError):
                _ = t[op1 - op2].show()
            with pytest.raises(exc.OperationalError):
                _ = t[op1 * op2].show()


    def test_inline_dict(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        df = t[[{'a': t.c1, 'b': {'c': t.c2}, 'd': 1, 'e': {'f': 2}}]]
        result = df.show()
        print(result)

    def test_inline_array(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        result = t[[ [[t.c2, 1], [t.c2, 2]] ]].show()
        t = result.col_types[0]
        assert t.is_array_type()
        assert isinstance(t, ArrayType)
        assert t.shape == (2, 2)
        assert t.dtype == ColumnType.Type.INT

    def test_json_mapper(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        # top-level is dict
        df = t[t.c6.f5['*'] >> (R + 1)]
        res = df.show()
        print(res)
        _ = t[t.c7['*'].f5 >> [R[3], R[2], R[1], R[0]]]
        _ = _.show()
        print(_)
        # target expr contains global-scope dependency
        df = t[
            t.c6.f5['*'] >> (R * t.c6.f5[1])
        ]
        res = df.show()
        print(res)

    def test_dicts(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        # top-level is dict
        _ = t[t.c6.f1]
        _ = _.show()
        print(_)
        #_ = t[t.c6.f2].show()
        #_ = t[t.c6.f5].show()
        _ = t[t.c6.f6.f8].show()
        _ = t[cast(t.c6.f6.f8, ArrayType((4,), ColumnType.Type.FLOAT))].show()

        # top-level is array
        #_ = t[t.c7['*'].f1].show()
        #_ = t[t.c7['*'].f2].show()
        #_ = t[t.c7['*'].f5].show()
        _ = t[t.c7['*'].f6.f8].show()
        _ = t[cast(t.c7['*'].f6.f8, ArrayType((2, 4), ColumnType.Type.FLOAT))].show()
        print(_)

    def test_arrays(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        t.add_column(catalog.Column('array_col', computed_with=[[t.c2, 1], [1, t.c2]]))
        _ = t[t.array_col].show()
        print(_)
        _ = t[t.array_col[:, 0]].show()
        print(_)

    def test_select_list(self, img_tbl) -> None:
        t = img_tbl
        result = t[t.img].show(n=100)
        _ = result._repr_html_()
        df = t[t.img, udf_call(lambda img: img.rotate(60), ImageType(), tbl=t)]
        _ = df.show(n=100)._repr_html_()
        df = t[[t.img, t.img.rotate(60)]]
        _ = df.show(n=100)._repr_html_()

        with pytest.raises(exc.OperationalError):
            _ = t[t.img.rotate]

    def test_img_members(self, img_tbl) -> None:
        t = img_tbl
        result = t[t.img.crop((10, 10, 60, 60))].show(n=100)
        result = t[t.img.crop((10, 10, 60, 60)).resize((100, 100))].show(n=100)
        result = t[t.img.crop((10, 10, 60, 60)).resize((100, 100)).convert('L')].show(n=100)
        result = t[t.img.getextrema()].show(n=100)
        result = t[t.img, t.img.height, t.img.rotate(90)].show(n=100)
        _ = result._repr_html_()

    def test_img_functions(self, img_tbl) -> None:
        t = img_tbl
        result = t[blend(t.img, t.img.rotate(90), 0.5)].show(100)
        print(result)
        result = t[encode_image(t.img)].show(10)
        print(result)
        _ = result._repr_html_()
        _ = t.img.entropy() > 1
        _ = _.extract_sql_predicate()
        _ = (t.img.entropy() > 1) & (t.split == 'train')
        _ = _.extract_sql_predicate()
        _ = (t.img.entropy() > 1) & (t.split == 'train') & (t.split == 'val')
        _ = _.extract_sql_predicate()
        _ = (t.split == 'train') & (t.img.entropy() > 1) & (t.split == 'val') & (t.img.entropy() < 0)
        _ = _.extract_sql_predicate()
        _ = t[(t.split == 'train') & (t.category == 'n03445777')][t.img].show()
        print(_)
        result = t[t.img.width > 1].show()
        print(result)
        result = t[(t.split == 'val') & (t.img.entropy() > 1) & (t.category == 'n03445777')].show()
        print(result)
        result = t[
            (t.split == 'train') & (t.img.entropy() > 1) & (t.split == 'val') & (t.img.entropy() < 0)
        ][t.img, t.split].show()
        print(result)

    def test_categoricals_map(self, img_tbl) -> None:
        t = img_tbl
        m = t[t.category].categorical_map()
        _ = t[dict_map(t.category, m)].show()
        print(_)

    def test_similarity(self, indexed_img_tbl: catalog.Table) -> None:
        t = indexed_img_tbl
        _ = t.show(30)
        probe = t[t.img, t.category].show(1)
        img = probe[0, 0]
        result = t[t.img.nearest(img)].show(10)
        assert len(result) == 10
        # nearest() with one SQL predicate and one Python predicate
        result = t[t.img.nearest(img) & (t.category == probe[0, 1]) & (t.img.width > 1)].show(10)
        # TODO: figure out how to verify results
        #assert len(result) == 3

        result = t[t.img.matches('musical instrument')].show(10)
        assert len(result) == 10
        # matches() with one SQL predicate and one Python predicate
        french_horn_category = 'n03394916'
        result = t[
            t.img.matches('musical instrument') & (t.category == french_horn_category) & (t.img.width > 1)
        ].show(10)
        #assert len(result) == 6

    # TODO: this doesn't work when combined with test_similarity(), for some reason the data table for img_tbl
    # doesn't get created; why?
    def test_similarity2(self, img_tbl: catalog.Table) -> None:
        t = img_tbl
        probe = t[t.img].show(1)
        img = probe[0, 0]

        with pytest.raises(exc.OperationalError):
            _ = t[t.img.nearest(img)].show(10)
        with pytest.raises(exc.OperationalError):
            _ = t[t.img.matches('musical instrument')].show(10)

    def test_serialization(self, test_tbl: catalog.Table, img_tbl: catalog.Table) -> None:
        t = test_tbl
        # add array column
        t.add_column(catalog.Column('c8', computed_with=[[1, 2, 3], [4, 5, 6]]))
        img_t = img_tbl
        test_exprs = [
            t.c1,
            t.c7['*'].f1,
            Literal('test'),
            InlineDict({
                'a': t.c1, 'b': t.c6.f1, 'c': 17,
                'd': InlineDict({'e': t.c2}),
                'f': InlineArray((t.c3, t.c3))
            }),
            InlineArray([[t.c2, t.c2], [t.c2, t.c2]]),
            t.c2 > 5,
            ~(t.c2 > 5),
            (t.c2 > 5) & (t.c1 == 'test'),
            (t.c2 > 5) | (t.c1 == 'test'),
            t.c7['*'].f5 >> [R[3], R[2], R[1], R[0]],
            t.c8[0, 1:],
            utils.sum_uda(t.c2).window(partition_by=t.c4, order_by=t.c3),
        ]
        for e in test_exprs:
            e_serialized = e.serialize()
            e_deserialized = Expr.deserialize(e_serialized, t)
            assert e.equals(e_deserialized)

        img_test_exprs = [
            img_t.img.width,
            img_t.img.rotate(90),
        ]
        for e in img_test_exprs:
            e_serialized = e.serialize()
            e_deserialized = Expr.deserialize(e_serialized, img_t)
            assert e.equals(e_deserialized)

    def test_subexprs(self, img_tbl: catalog.Table) -> None:
        t = img_tbl
        e = t.img
        subexprs = [s for s in e.subexprs()]
        assert len(subexprs) == 1
        e = t.img.rotate(90).resize((224, 224))
        subexprs = [s for s in e.subexprs()]
        assert len(subexprs) == 3
        subexprs = [s for s in e.subexprs() if isinstance(s, ColumnRef)]
        assert len(subexprs) == 1
        assert t.img.equals(subexprs[0])

    def test_window_fns(self, test_db: catalog.Db, test_tbl: catalog.Table) -> None:
        db = test_db
        t = test_tbl
        _ = t[utils.sum_uda(t.c2).window(partition_by=t.c4, order_by=t.c3)].show(100)
        print(_)
        # backfill works
        t.add_column(catalog.Column('c9', computed_with=utils.sum_uda(t.c2).window(partition_by=t.c4, order_by=t.c3)))

        c2 = catalog.Column('c2', IntType(), nullable=False)
        c3 = catalog.Column('c3', FloatType(), nullable=False)
        c4 = catalog.Column('c4', BoolType(), nullable=False)
        new_t = db.create_table('insert_test', [c2, c3, c4])
        new_t.add_column(catalog.Column(
            'c2_sum', computed_with=utils.sum_uda(new_t.c2).window(partition_by=new_t.c4, order_by=new_t.c3)))
        data_df = t[t.c2, t.c4, t.c3].show(0).to_pandas()
        new_t.insert_pandas(data_df)
        _ = new_t.show(0)
        print(_)

    def test_aggregates(self, test_tbl: catalog.Table) -> None:
        t = test_tbl
        _ = t[t.c2 % 2, sum(t.c2), count(t.c2), sum(t.c2) + count(t.c2), sum(t.c2) + t.c2 % 2].group_by(t.c2 % 2).show()
        print(_)



from typing import List, Set
import PIL.Image
import hnswlib
import numpy as np

from pixeltable.env import Env

class VectorIndex:
    def __init__(self, name: str, dim: int, idx: hnswlib.Index):
        self.name = name
        self.dim = dim
        self.idx = idx

    @classmethod
    def create(cls, name: str, dim: int) -> 'VectorIndex':
        idx = hnswlib.Index(space='cosine', dim=dim)
        idx.init_index(max_elements=1000, M=64, ef_construction=200)
        filename = cls._filename(name)
        idx.save_index(filename)
        return VectorIndex(name, dim, idx)

    @classmethod
    def load(cls, name: str, dim: int) -> 'VectorIndex':
        idx = hnswlib.Index(space='cosine', dim=dim)
        filename = cls._filename(name)
        idx.load_index(filename)
        return VectorIndex(name, dim, idx)

    def insert(self, data: np.ndarray, rowids: np.ndarray) -> None:
        assert data.shape[0] == rowids.shape[0]
        total = self.idx.element_count + data.shape[0]
        if total > self.idx.max_elements:
            self.idx.resize_index(int(total * 1.1))
        self.idx.add_items(data, rowids)
        filename = self._filename(self.name)
        self.idx.save_index(filename)

    def search(self, embed: np.ndarray, num_nn: int, valid_rowids: Set[int]) -> List[int]:
        """
        Returns rowids of k nearest neighbors.
        """
        assert embed.shape == (512,)
        k = num_nn
        while True:
            nn, _ = self.idx.knn_query(embed.reshape(1, 512), k)
            # tolist(): make sure result contains ints, not uint64
            result = [rowid for rowid in nn.squeeze().tolist() if rowid in valid_rowids]
            if len(result) >= num_nn:
                return list(result[:num_nn])
            if k >= self.idx.element_count:
                # nothing left to look for
                return result
            k *= 2

    @classmethod
    def _filename(cls, name: str) -> str:
        return str(Env.get().nnidx_dir / f'idx_{name}')


from typing import Callable, Any, Optional, Tuple
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
import tensorflow as tf


from pixeltable.type_system import ColumnType, ArrayType
import pixeltable.utils.tf as tf_utils
from pixeltable.function import Function

def TFModelFunction(
        model: tf.keras.Model, param_type: ColumnType, output_shape: Optional[Tuple[int, ...]] = None,
        preprocess: Optional[Callable] = None) -> Function:
    def predict(arg: Any) -> Any:
        tensor = tf_utils.to_tensor(arg, param_type)
        tensor = tf.expand_dims(tensor, axis=0)
        if preprocess is not None:
            tensor = preprocess(tensor)
        output = model.predict(tensor)
        return output[0]
    return Function(ArrayType(output_shape, dtype=ColumnType.DType.FLOAT32), [param_type], eval_fn=predict)


from typing import List
import ffmpeg
import glob
from pathlib import Path

from pixeltable.exceptions import OperationalError


def extract_frames(video_path_str: str, output_path_prefix: str, fps: int = 0) -> List[str]:
    """
    Extract frames at given fps as jpg files (fps == 0: all frames).
    Returns list of frame file paths.
    """
    video_path = Path(video_path_str)
    if not video_path.exists():
        raise OperationalError(f'File not found: {video_path_str}')
    if not video_path.is_file():
        raise OperationalError(f'Not a file: {video_path_str}')
    output_path_str = f'{output_path_prefix}_%07d.jpg'
    s = ffmpeg.input(video_path)
    if fps > 0:
        s = s.filter('fps', fps)
    s = s.output(output_path_str, loglevel='quiet')
    try:
        s.run()
    except ffmpeg.Error:
        raise OperationalError(f'ffmpeg exception')

    # collect generated files
    frame_paths = glob.glob(f'{output_path_prefix}_*.jpg')
    frame_paths.sort()
    return frame_paths



from typing import Any, Tuple, List, Generator
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
import tensorflow as tf

from pixeltable.dataframe import DataFrame
import pixeltable.utils.tf as tf_utils


def to_dataset(df: DataFrame) -> tf.data.Dataset:
    """
    Convert df result set to tf.data.Dataset.
    """
    result_set = df.show(0)
    # construct output signature
    if len(result_set.col_types) == 1 and result_set.col_types[0].is_json_type():
        # we're only getting one column back and it's a dict: make the dataset type a dict
        type_spec = result_set.col_types[0].to_tf()
    else:
        tf_types = [t.to_tf() for t in result_set.col_types]
        type_spec = tuple(tf_types)

    def f() -> Generator[Tuple[Any, ...], None, None]:
        if isinstance(type_spec, dict):
            json_type = result_set.col_types[0]
            for row in result_set.rows:
                res = {k: tf_utils.to_tensor(v, json_type.type_spec[k]) for k, v in row[0].items()}
                yield res
        else:
            for row in result_set.rows:
                tf_row = [tf_utils.to_tensor(val, result_set.col_types[i]) for i, val in enumerate(row)]
                yield tuple(tf_row)
    return tf.data.Dataset.from_generator(f, output_signature=type_spec)


import abc
import copy
import datetime
import enum
import inspect
import typing
from typing import Union, Optional, List, Callable, Any, Dict, Tuple, Set, Generator, Iterator
import operator
import json
import io

import PIL.Image
import jmespath
import numpy as np
import sqlalchemy as sql

from pixeltable import catalog
from pixeltable.type_system import \
    ColumnType, InvalidType, StringType, IntType, FloatType, BoolType, TimestampType, ImageType, JsonType, ArrayType
from pixeltable.function import Function
from pixeltable import exceptions as exc
from pixeltable.utils import clip

# Python types corresponding to our literal types
LiteralPythonTypes = Union[str, int, float, bool, datetime.datetime, datetime.date]


class ComparisonOperator(enum.Enum):
    LT = 0
    LE = 1
    EQ = 2
    NE = 3
    GT = 4
    GE = 5

    def __str__(self) -> str:
        if self == self.LT:
            return '<'
        if self == self.LE:
            return '<='
        if self == self.EQ:
            return '=='
        if self == self.GT:
            return '>'
        if self == self.GE:
            return '>='


class LogicalOperator(enum.Enum):
    AND = 0
    OR = 1
    NOT = 2


class ArithmeticOperator(enum.Enum):
    ADD = 0
    SUB = 1
    MUL = 2
    DIV = 3
    MOD = 4

    def __str__(self) -> str:
        if self == self.ADD:
            return '+'
        if self == self.SUB:
            return '-'
        if self == self.MUL:
            return '*'
        if self == self.DIV:
            return '/'
        if self == self.MOD:
            return '%'


class ExprScope:
    """
    Representation of the scope in which an Expr needs to be evaluated. Used to determine nesting of scopes.
    parent is None: outermost scope
    """
    def __init__(self, parent: Optional['ExprScope']):
        self.parent = parent

    def is_contained_in(self, other: 'ExprScope') -> bool:
        if self == other:
            return True
        if self.parent is None:
            return False
        return self.parent.is_contained_in(other)


_GLOBAL_SCOPE = ExprScope(None)


class Expr(abc.ABC):
    """
    Rules for using state in subclasses:
    - all state except for components and data/sql_row_idx is shared between copies of an Expr
    - data/sql_row_idx is set during analysis (DataFrame.show())
    - during eval(), components can only be accessed via self.components; any Exprs outside of that won't
      have data_row_idx set
    """
    def __init__(self, col_type: ColumnType):
        self.col_type = col_type
        # index of the expr's value in the data row; set for all materialized exprs; -1: invalid
        # not set for subexprs that don't need to be materialized because the parent can be materialized via SQL
        self.data_row_idx = -1
        # index of the expr's value in the SQL row; only set for exprs that can be materialized in SQL; -1: invalid
        self.sql_row_idx = -1
        self.components: List[Expr] = []  # the subexprs that are needed to construct this expr

    def dependencies(self) -> List['Expr']:
        """
        Returns all exprs that need to have been evaluated before eval() can be called on this one.
        """
        return self.components

    def scope(self) -> ExprScope:
        # by default this is the innermost scope of any of our components
        result = _GLOBAL_SCOPE
        for c in self.components:
            c_scope = c.scope()
            if c_scope.is_contained_in(result):
                result = c_scope
        return result

    def bind_rel_paths(self, mapper: Optional['JsonMapper'] = None) -> None:
        """
        Binds relative JsonPaths to mapper.
        This needs to be done in a separate phase after __init__(), because RelativeJsonPath()(-1) cannot be resolved
        by the immediately containing JsonMapper during initialization.
        """
        for c in self.components:
            c.bind_rel_paths(mapper)

    def display_name(self) -> str:
        """
        Displayed column name in DataFrame. '': assigned by DataFrame
        """
        return ''

    def equals(self, other: 'Expr') -> bool:
        """
        Subclass-specific comparison. Implemented as a function because __eq__() is needed to construct Comparisons.
        """
        if type(self) != type(other):
            return False
        if len(self.components) != len(other.components):
            return False
        for i in range(len(self.components)):
            if not self.components[i].equals(other.components[i]):
                return False
        return self._equals(other)

    @classmethod
    def list_equals(cls, a: List['Expr'], b: List['Expr']) -> bool:
        if len(a) != len(b):
            return False
        for i in range(len(a)):
            if not a[i].equals(b[i]):
                return False
        return True

    def copy(self) -> 'Expr':
        """
        Creates a copy that can be evaluated separately: it doesn't share any eval context (data/sql_row_idx)
        but shares everything else (catalog objects, etc.)
        """
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        result.data_row_idx = -1
        result.sql_row_idx = -1
        result.components = [c.copy() for c in self.components]
        return result

    @classmethod
    def copy_list(cls, expr_list: List['Expr']) -> List['Expr']:
        return [e.copy() for e in expr_list]

    def __deepcopy__(self, memo={}) -> 'Expr':
        # we don't need to create an actual deep copy because all state other than execution state is read-only
        result = self.copy()
        memo[id(self)] = result
        return result

    def subexprs(self) -> Generator['Expr', None, None]:
        """
        Iterate over all subexprs, including self.
        """
        for c in self.components:
            yield from c.subexprs()
        yield self

    @classmethod
    def list_subexprs(cls, expr_list: List['Expr']) -> Generator['Expr', None, None]:
        """
        Produce subexprs for all exprs in list.
        """
        for e in expr_list:
            yield from e.subexprs()

    @classmethod
    def from_object(cls, o: object) -> Optional['Expr']:
        """
        Try to turn a literal object into an Expr.
        """
        if isinstance(o, Expr):
            return o
        if isinstance(o, dict):
            return InlineDict(o)
        elif isinstance(o, list):
            return InlineArray(tuple(o))
        return None

    @abc.abstractmethod
    def _equals(self, other: 'Expr') -> bool:
        pass

    @abc.abstractmethod
    def sql_expr(self) -> Optional[sql.sql.expression.ClauseElement]:
        """
        If this expr can be materialized directly in SQL:
        - returns a ClauseElement
        - eval() will not be called (exception: Literal)
        Otherwise
        - returns None
        - eval() will be called
        """
        pass

    @abc.abstractmethod
    def eval(self, data_row: List[Any]) -> None:
        """
        Compute the expr value for data_row and store the result in data_row[data_row_idx].
        Not called if sql_expr() != None (exception: Literal).
        """
        pass

    def serialize(self) -> str:
        return json.dumps(self.as_dict())

    def as_dict(self) -> Dict:
        """
        Turn Expr object into a dict that can be passed to json.dumps().
        Subclasses override _as_dict().
        """
        return {
            '_classname': self.__class__.__name__,
            **self._as_dict(),
        }

    @classmethod
    def as_dict_list(self, expr_list: List['Expr']) -> List[Dict]:
        return [e.as_dict() for e in expr_list]

    def _as_dict(self) -> Dict:
        if len(self.components) > 0:
            return {'components': [c.as_dict() for c in self.components]}
        return {}

    @classmethod
    def deserialize(cls, dict_str: str, t: catalog.Table) -> 'Expr':
        return cls.from_dict(json.loads(dict_str), t)

    @classmethod
    def from_dict(cls, d: Dict, t: catalog.Table) -> 'Expr':
        """
        Turn dict that was produced by calling Expr.as_dict() into an instance of the correct Expr subclass.
        """
        assert '_classname' in d
        type_class = globals()[d['_classname']]
        components: List[Expr] = []
        if 'components' in d:
            components = [cls.from_dict(component_dict, t) for component_dict in d['components']]
        return type_class._from_dict(d, components, t)

    @classmethod
    def from_dict_list(cls, dict_list: List[Dict], t: catalog.Table) -> List['Expr']:
        return [cls.from_dict(d, t) for d in dict_list]

    @classmethod
    def _from_dict(cls, d: Dict, components: List['Expr'], t: catalog.Table) -> 'Expr':
        assert False, 'not implemented'

    def __getitem__(self, index: object) -> 'Expr':
        if self.col_type.is_json_type():
            return JsonPath(self).__getitem__(index)
        if self.col_type.is_array_type():
            return ArraySlice(self, index)
        raise exc.Error(f'Type {self.col_type} is not subscriptable')

    def __getattr__(self, name: str) -> 'ImageMemberAccess':
        """
        ex.: <img col>.rotate(60)
        """
        if not self.col_type.is_image_type():
            raise exc.OperationalError(f'Member access not supported on type {self.col_type}: {name}')
        return ImageMemberAccess(name, self)

    def __lt__(self, other: object) -> 'Comparison':
        return self._make_comparison(ComparisonOperator.LT, other)

    def __le__(self, other: object) -> 'Comparison':
        return self._make_comparison(ComparisonOperator.LE, other)

    def __eq__(self, other: object) -> 'Comparison':
        return self._make_comparison(ComparisonOperator.EQ, other)

    def __ne__(self, other: object) -> 'Comparison':
        return self._make_comparison(ComparisonOperator.NE, other)

    def __gt__(self, other: object) -> 'Comparison':
        return self._make_comparison(ComparisonOperator.GT, other)

    def __ge__(self, other: object) -> 'Comparison':
        return self._make_comparison(ComparisonOperator.GE, other)

    def _make_comparison(self, op: ComparisonOperator, other: object) -> 'Comparison':
        """
        other: Union[Expr, LiteralPythonTypes]
        """
        # TODO: check for compatibility
        if isinstance(other, Expr):
            return Comparison(op, self, other)
        if isinstance(other, typing.get_args(LiteralPythonTypes)):
            return Comparison(op, self, Literal(other))  # type: ignore[arg-type]
        raise TypeError(f'Other must be Expr or literal: {type(other)}')

    def __add__(self, other: object) -> 'ArithmeticExpr':
        return self._make_arithmetic_expr(ArithmeticOperator.ADD, other)

    def __sub__(self, other: object) -> 'ArithmeticExpr':
        return self._make_arithmetic_expr(ArithmeticOperator.SUB, other)

    def __mul__(self, other: object) -> 'ArithmeticExpr':
        return self._make_arithmetic_expr(ArithmeticOperator.MUL, other)

    def __truediv__(self, other: object) -> 'ArithmeticExpr':
        return self._make_arithmetic_expr(ArithmeticOperator.DIV, other)

    def __mod__(self, other: object) -> 'ArithmeticExpr':
        return self._make_arithmetic_expr(ArithmeticOperator.MOD, other)

    def _make_arithmetic_expr(self, op: ArithmeticOperator, other: object) -> 'ArithmeticExpr':
        """
        other: Union[Expr, LiteralPythonTypes]
        """
        # TODO: check for compatibility
        if isinstance(other, Expr):
            return ArithmeticExpr(op, self, other)
        if isinstance(other, typing.get_args(LiteralPythonTypes)):
            return ArithmeticExpr(op, self, Literal(other))  # type: ignore[arg-type]
        raise TypeError(f'Other must be Expr or literal: {type(other)}')


class ColumnRef(Expr):
    def __init__(self, col: catalog.Column):
        super().__init__(col.col_type)
        self.col = col

    def __getattr__(self, name: str) -> Expr:
        if self.col_type.is_json_type():
            return JsonPath(self).__getattr__(name)
        return super().__getattr__(name)

    def display_name(self) -> str:
        return self.col.name

    def _equals(self, other: 'ColumnRef') -> bool:
        return self.col == other.col

    def sql_expr(self) -> Optional[sql.sql.expression.ClauseElement]:
        return self.col.sa_col

    def eval(self, data_row: List[Any]) -> None:
        # we get called while materializing computed cols
        pass

    def _as_dict(self) -> Dict:
        return {'col_id': self.col.id}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr], t: catalog.Table) -> 'Expr':
        assert 'col_id' in d
        return cls(t.cols_by_id[d['col_id']])


class FunctionCall(Expr):
    def __init__(self, fn: Function, args: Tuple[Any] = None):
        super().__init__(fn.return_type)
        self.fn = fn

        if fn.param_types is not None:
            # check if arg types match param types and convert values, if necessary
            if len(args) != len(fn.param_types):
                raise exc.OperationalError(
                    f"Number of arguments doesn't match parameter list: {args} vs {fn.param_types}")
            args = list(args)
            for i in range(len(args)):
                if not isinstance(args[i], Expr):
                    # TODO: check non-Expr args
                    continue
                if args[i].col_type == fn.param_types[i]:
                    # nothing to do
                    continue
                converter = args[i].col_type.conversion_fn(fn.param_types[i])
                if converter is None:
                    raise exc.OperationalError(f'Cannot convert {args[i].col_type} to {fn.param_types[i]}')
                if converter == ColumnType.no_conversion:
                    # nothing to do
                    continue
                convert_fn = Function(fn.param_types[i], [args[i].col_type], eval_fn=converter)
                args[i] = FunctionCall(convert_fn, (args[i],))

        self.components = [arg for arg in args if isinstance(arg, Expr)]
        self.args = [arg if not isinstance(arg, Expr) else None for arg in args]

        # window function state
        self.partition_by_idx = -1  # self.components[self.pb_index:] contains partition_by exprs
        self.order_by: List[Expr] = []
        # execution state for window functions
        self.aggregator: Optional[Any] = self.fn.init_fn() if self.fn.is_aggregate else None
        self.current_partition_vals: Optional[List[Any]] = None

    @property
    def _eval_fn(self) -> Optional[Callable]:
        return self.fn.eval_fn

    def _equals(self, other: 'FunctionCall') -> bool:
        if self.fn != other.fn:
            return False
        if len(self.args) != len(other.args):
            return False
        for i in range(len(self.args)):
            if self.args[i] != other.args[i]:
                return False
        if self.partition_by_idx != other.partition_by_idx:
            return False
        if not self.list_equals(self.order_by, other.order_by):
            return False
        return True

    def window(
            self, partition_by: Optional[Union[Expr, List[Expr]]] = None,
            order_by: Optional[Union[Expr, List[Expr]]] = None
    ) -> 'FunctionCall':
        if not self.fn.is_aggregate:
            raise exc.Error(f'The window() clause is only allowed for aggregate functions')
        if partition_by is None and order_by is None:
            raise exc.Error('The window() clause requires at least one parameter not to be None')
        if partition_by is not None and not isinstance(partition_by, list):
            partition_by = [partition_by]
        if order_by is not None:
            self.order_by = order_by if isinstance(order_by, list) else [order_by]
        # we only need to record the partition_by exprs in self.components, because the order_by values aren't
        # used during evaluation (the SQL store will return rows in that order)
        if partition_by is not None:
            self.partition_by_idx = len(self.components)
            self.components.extend(partition_by)
        return self

    @property
    def partition_by(self) -> List[Expr]:
        if self.partition_by_idx == -1:
            return []
        return self.components[self.partition_by_idx:]

    @property
    def is_window_fn_call(self) -> bool:
        return self.fn.is_aggregate and (self.partition_by_idx != -1 or len(self.order_by) > 0)

    def get_window_sort_exprs(self) -> List[Expr]:
        return [*self.partition_by, *self.order_by]

    @property
    def is_agg_fn_call(self) -> bool:
        return self.fn.is_aggregate and self.partition_by_idx == -1 and len(self.order_by) == 0

    def get_agg_order_by(self) -> List[Expr]:
        assert self.is_agg_fn_call
        result: List[Expr] = []
        component_idx = 0
        for arg_idx in range(len(self.args)):
            if arg_idx in self.fn.order_by:
                assert self.args[arg_idx] is None  # this is an Expr, not something else
                result.append(self.components[component_idx])
            if self.args[arg_idx] is None:
                component_idx += 1
        return result

    def sql_expr(self) -> Optional[sql.sql.expression.ClauseElement]:
        # TODO: implement for standard aggregate functions
        return None

    def reset_agg(self) -> None:
        """
        Init agg state
        """
        assert self.is_agg_fn_call
        self.aggregator = self.fn.init_fn()

    def update(self, data_row: List[Any]) -> None:
        """
        Update agg state
        """
        assert self.is_agg_fn_call
        args = self._make_args(data_row)
        self.fn.update_fn(self.aggregator, *args)

    def _make_args(self, data_row: List[Any]) -> List[Any]:
        args = copy.copy(self.args)
        # fill in missing child values
        i = 0
        for j in range(len(args)):
            if args[j] is None:
                args[j] = data_row[self.components[i].data_row_idx]
                i += 1
        return args

    def eval(self, data_row: List[Any]) -> None:
        args = self._make_args(data_row)
        if not self.fn.is_aggregate:
            data_row[self.data_row_idx] = self.fn.eval_fn(*args)
        elif self.is_window_fn_call:
            if self.partition_by_idx != -1:
                if self.current_partition_vals is None:
                    self.current_partition_vals = [None] * len(self.partition_by)
                partition_vals = [data_row[e.data_row_idx] for e in self.partition_by]
                if partition_vals != self.current_partition_vals:
                    # new partition
                    self.aggregator = self.fn.init_fn()
                    self.current_partition_vals = partition_vals
            elif self.aggregator is None:
                self.aggregator = self.fn.init_fn()
            self.fn.update_fn(self.aggregator, *args)
            data_row[self.data_row_idx] = self.fn.value_fn(self.aggregator)
        else:
            assert self.is_agg_fn_call
            data_row[self.data_row_idx] = self.fn.value_fn(self.aggregator)

    def _as_dict(self) -> Dict:
        result = {'fn': self.fn.as_dict(), 'args': self.args, **super()._as_dict()}
        if self.fn.is_aggregate:
            result.update({'partition_by_idx': self.partition_by_idx, 'order_by': Expr.as_dict_list(self.order_by)})
        return result

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr], t: catalog.Table) -> 'Expr':
        assert 'fn' in d
        assert 'args' in d
        # reassemble args
        args = [arg if arg is not None else components[i] for i, arg in enumerate(d['args'])]
        fn_call = cls(Function.from_dict(d['fn']), args)
        if fn_call.fn.is_aggregate:
            fn_call.partition_by_idx = d['partition_by_idx']
            fn_call.components.extend(components[fn_call.partition_by_idx:])
            fn_call.order_by = Expr.from_dict_list(d['order_by'], t)
        return fn_call


def _caller_return_type(caller: Expr, *args: object, **kwargs: object) -> ColumnType:
    return caller.col_type

def _convert_return_type(caller: Expr, *args: object, **kwargs: object) -> ColumnType:
    mode_str = args[0]
    assert isinstance(mode_str, str)
    assert isinstance(caller.col_type, ImageType)
    return ImageType(
        width=caller.col_type.width, height=caller.col_type.height, mode=ImageType.Mode.from_pil(mode_str))

def _crop_return_type(caller: Expr, *args: object, **kwargs: object) -> ColumnType:
    left, upper, right, lower = args[0]
    assert isinstance(caller.col_type, ImageType)
    return ImageType(width=(right - left), height=(lower - upper), mode=caller.col_type.mode)

def _resize_return_type(caller: Expr, *args: object, **kwargs: object) -> ColumnType:
    w, h = args[0]
    assert isinstance(caller.col_type, ImageType)
    return ImageType(width=w, height=h, mode=caller.col_type.mode)

# This only includes methods that return something that can be displayed in pixeltable
# and that make sense to call (counterexample: copy() doesn't make sense to call)
# This is hardcoded here instead of being dynamically extracted from the PIL type stubs because
# doing that is messy and it's unclear whether it has any advantages.
# TODO: how to capture return values like List[Tuple[int, int]]?
# dict from method name to (function to compute value, function to compute return type)
# TODO: JsonTypes() where it should be ArrayType(): need to determine the shape and base type
_PIL_METHOD_INFO: Dict[str, Union[ColumnType, Callable]] = {
    'convert': _convert_return_type,
    'crop': _crop_return_type,
    'effect_spread': _caller_return_type,
    'entropy': FloatType(),
    'filter': _caller_return_type,
    'getbands': ArrayType((None,), ColumnType.Type.STRING),
    'getbbox': ArrayType((4,), ColumnType.Type.INT),
    'getchannel': _caller_return_type,
    'getcolors': JsonType(),
    'getextrema': JsonType(),
    'getpalette': JsonType(),
    'getpixel': JsonType(),
    'getprojection': JsonType(),
    'histogram': JsonType(),
# TODO: what to do with this? it modifies the img in-place
#    paste: <ast.Constant object at 0x7f9e9a9be3a0>
    'point': _caller_return_type,
    'quantize': _caller_return_type,
    'reduce': _caller_return_type,  # TODO: this is incorrect
    'remap_palette': _caller_return_type,
    'resize': _resize_return_type,
    'rotate': _caller_return_type,  # TODO: this is incorrect
# TODO: this returns a Tuple[Image], which we can't express
#    split: <ast.Subscript object at 0x7f9e9a9cc9d0>
    'transform': _caller_return_type,  # TODO: this is incorrect
    'transpose': _caller_return_type,  # TODO: this is incorrect
}


# TODO: this doesn't dig up all attrs for actual jpeg images
def _create_pil_attr_info() -> Dict[str, ColumnType]:
    # create random Image to inspect for attrs
    img = PIL.Image.new('RGB', (100, 100))
    # we're only interested in public attrs (including properties)
    result: Dict[str, ColumnType] = {}
    for name in [name for name in dir(img) if not callable(getattr(img, name)) and not name.startswith('_')]:
        if getattr(img, name) is None:
            continue
        if isinstance(getattr(img, name), str):
            result[name] = StringType()
        if isinstance(getattr(img, name), int):
            result[name] = IntType()
        if getattr(img, name) is dict:
            result[name] = JsonType()
    return result


class ImageMemberAccess(Expr):
    """
    Access of either an attribute or function member of PIL.Image.Image.
    Ex.: tbl.img_col_ref.rotate(90), tbl.img_col_ref.width
    """
    attr_info = _create_pil_attr_info()
    special_img_predicates = ['nearest', 'matches']

    def __init__(self, member_name: str, caller: Expr):
        if member_name in self.special_img_predicates:
            super().__init__(InvalidType())  # requires FunctionCall to return value
        elif member_name in _PIL_METHOD_INFO:
            super().__init__(InvalidType())  # requires FunctionCall to return value
        elif member_name in self.attr_info:
            super().__init__(self.attr_info[member_name])
        else:
            raise exc.OperationalError(f'Unknown Image member: {member_name}')
        self.member_name = member_name
        self.components = [caller]

    def display_name(self) -> str:
        return self.member_name

    @property
    def _caller(self) -> Expr:
        return self.components[0]

    def _as_dict(self) -> Dict:
        return {'member_name': self.member_name, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr], t: catalog.Table) -> Expr:
        assert 'member_name' in d
        assert len(components) == 1
        return cls(d['member_name'], components[0])

    # TODO: correct signature?
    def __call__(self, *args, **kwargs) -> Union['ImageMethodCall', 'ImageSimilarityPredicate']:
        caller = self._caller
        call_signature = f'({",".join([type(arg).__name__ for arg in args])})'
        if self.member_name == 'nearest':
            # - caller must be ColumnRef
            # - signature is (PIL.Image.Image)
            if not isinstance(caller, ColumnRef):
                raise exc.OperationalError(f'nearest(): caller must be an IMAGE column')
            if len(args) != 1 or not isinstance(args[0], PIL.Image.Image):
                raise exc.OperationalError(
                    f'nearest(): required signature is (PIL.Image.Image) (passed: {call_signature})')
            return ImageSimilarityPredicate(caller, img=args[0])

        if self.member_name == 'matches':
            # - caller must be ColumnRef
            # - signature is (str)
            if not isinstance(caller, ColumnRef):
                raise exc.OperationalError(f'matches(): caller must be an IMAGE column')
            if len(args) != 1 or not isinstance(args[0], str):
                raise exc.OperationalError(f"matches(): required signature is (str) (passed: {call_signature})")
            return ImageSimilarityPredicate(caller, text=args[0])

        # TODO: verify signature
        return ImageMethodCall(self.member_name, caller, *args, **kwargs)

    def _equals(self, other: 'ImageMemberAccess') -> bool:
        return self.member_name == other.member_name

    def sql_expr(self) -> Optional[sql.sql.expression.ClauseElement]:
        return None

    def eval(self, data_row: List[Any]) -> None:
        caller_val = data_row[self._caller.data_row_idx]
        try:
            data_row[self.data_row_idx] = getattr(caller_val, self.member_name)
        except AttributeError:
            data_row[self.data_row_idx] = None


class ImageMethodCall(FunctionCall):
    """
    Ex.: tbl.img_col_ref.rotate(90)
    """
    def __init__(self, method_name: str, caller: Expr, *args: object, **kwargs: object):
        assert method_name in _PIL_METHOD_INFO
        self.method_name = method_name
        method_info = _PIL_METHOD_INFO[self.method_name]
        if isinstance(method_info, ColumnType):
            return_type = method_info
        else:
            return_type = method_info(caller, *args, **kwargs)
        # TODO: register correct parameters
        fn = Function(return_type, None, module_name='PIL.Image', eval_symbol=f'Image.{method_name}')
        super().__init__(fn, (caller, *args))
        # TODO: deal with kwargs

    def display_name(self) -> str:
        return self.method_name

    def _as_dict(self) -> Dict:
        return {'method_name': self.method_name, 'args': self.args, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr], t: catalog.Table) -> 'Expr':
        """
        We're implementing this, instead of letting FunctionCall handle it, in order to return an
        ImageMethodCall instead of a FunctionCall, which is useful for testing that a serialize()/deserialize()
        roundtrip ends up with the same Expr.
        """
        assert 'method_name' in d
        assert 'args' in d
        # reassemble args, but skip args[0], which is our caller
        args = [arg if arg is not None else components[i+1] for i, arg in enumerate(d['args'][1:])]
        return cls(d['method_name'], components[0], *args)


class JsonPath(Expr):
    def __init__(self, anchor: Optional[ColumnRef], path_elements: List[str] = [], scope_idx: int = 0):
        """
        anchor can be None, in which case this is a relative JsonPath and the anchor is set later via set_anchor().
        scope_idx: for relative paths, index of referenced JsonMapper
        (0: indicates the immediately preceding JsonMapper, -1: the parent of the immediately preceding mapper, ...)
        """
        super().__init__(JsonType())
        if anchor is not None:
            self.components = [anchor]
        self.path_elements: List[Union[str, int]] = path_elements
        self.compiled_path = jmespath.compile(self._json_path()) if len(path_elements) > 0 else None
        self.scope_idx = scope_idx

    def _as_dict(self) -> Dict:
        return {'path_elements': self.path_elements, 'scope_idx': self.scope_idx, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr], t: catalog.Table) -> Expr:
        assert 'path_elements' in d
        assert 'scope_idx' in d
        assert len(components) <= 1
        anchor = components[0] if len(components) == 1 else None
        return cls(anchor, d['path_elements'], d['scope_idx'])

    @property
    def _anchor(self) -> Optional[Expr]:
        return None if len(self.components) == 0 else self.components[0]

    def set_anchor(self, anchor: Expr) -> None:
        assert len(self.components) == 0
        self.components = [anchor]

    def is_relative_path(self) -> bool:
        return self._anchor is None

    def bind_rel_paths(self, mapper: Optional['JsonMapper'] = None) -> None:
        if not self.is_relative_path():
            return
        # TODO: take scope_idx into account
        self.set_anchor(mapper.scope_anchor)

    def __call__(self, *args: object, **kwargs: object) -> 'JsonPath':
        """
        Construct a relative path that references an ancestor of the immediately enclosing JsonMapper.
        """
        if not self.is_relative_path():
            raise exc.OperationalError(f'() for an absolute path is invalid')
        if len(args) != 1 or not isinstance(args[0], int) or args[0] >= 0:
            raise exc.OperationalError(f'R() requires a negative index')
        return JsonPath(None, [], args[0])

    def __getattr__(self, name: str) -> 'JsonPath':
        assert isinstance(name, str)
        return JsonPath(self._anchor, self.path_elements + [name])

    def __getitem__(self, index: object) -> 'JsonPath':
        if isinstance(index, str) and index != '*':
            raise exc.OperationalError(f'Invalid json list index: {index}')
        return JsonPath(self._anchor, self.path_elements + [index])

    def __rshift__(self, other: object) -> 'JsonMapper':
        rhs_expr = Expr.from_object(other)
        if rhs_expr is None:
            raise exc.OperationalError(f'>> requires an expression on the right-hand side, found {type(other)}')
        return JsonMapper(self, rhs_expr)

    def display_name(self) -> str:
        anchor_name = self._anchor.display_name() if self._anchor is not None else ''
        return f'{anchor_name}.{self._json_path()}'

    def _equals(self, other: 'JsonPath') -> bool:
        return self.path_elements == other.path_elements

    def sql_expr(self) -> Optional[sql.sql.expression.ClauseElement]:
        """
        Postgres appears to have a bug: jsonb_path_query('{a: [{b: 0}, {b: 1}]}', '$.a.b') returns
        *two* rows (each containing col val 0), not a single row with [0, 0].
        We need to use a workaround: retrieve the entire dict, then use jmespath to extract the path correctly.
        """
        #path_str = '$.' + '.'.join(self.path_elements)
        #assert isinstance(self._anchor(), ColumnRef)
        #return sql.func.jsonb_path_query(self._anchor().col.sa_col, path_str)
        return None

    def _json_path(self) -> str:
        assert len(self.path_elements) > 0
        result: List[str] = []
        for element in self.path_elements:
            if element == '*':
                result.append('[*]')
            elif isinstance(element, str):
                result.append(f'{"." if len(result) > 0 else ""}{element}')
            elif isinstance(element, int):
                result.append(f'[{element}]')
        return ''.join(result)

    def eval(self, data_row: List[Any]) -> None:
        val = data_row[self._anchor.data_row_idx]
        if self.compiled_path is not None:
            val = self.compiled_path.search(val)
        data_row[self.data_row_idx] = val


RELATIVE_PATH_ROOT = JsonPath(None)


class Literal(Expr):
    def __init__(self, val: LiteralPythonTypes):
        if isinstance(val, str):
            super().__init__(StringType())
        if isinstance(val, int):
            super().__init__(IntType())
        if isinstance(val, float):
            super().__init__(FloatType())
        if isinstance(val, bool):
            super().__init__(BoolType())
        if isinstance(val, datetime.datetime) or isinstance(val, datetime.date):
            super().__init__(TimestampType())
        self.val = val

    def display_name(self) -> str:
        return 'Literal'

    def _equals(self, other: 'Literal') -> bool:
        return self.val == other.val

    def sql_expr(self) -> Optional[sql.sql.expression.ClauseElement]:
        # we need to return something here so that we can generate a Where clause for predicates
        # that involve literals (like Where c > 0)
        return sql.sql.expression.literal(self.val)

    def eval(self, data_row: List[Any]) -> None:
        # this will be called, even though sql_expr() does not return None
        data_row[self.data_row_idx] = self.val

    def _as_dict(self) -> Dict:
        return {'val': self.val, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr], t: catalog.Table) -> Expr:
        assert 'val' in d
        return cls(d['val'])


class InlineDict(Expr):
    """
    Dictionary 'literal' which can use Exprs as values.
    """
    def __init__(self, d: Dict):
        super().__init__(JsonType())  # we need to call this in order to populate self.components
        # dict_items contains
        # - for Expr fields: (key, index into components, None)
        # - for non-Expr fields: (key, -1, value)
        self.dict_items: List[Tuple[str, int, Any]] = []
        for key, val in d.items():
            if not isinstance(key, str):
                raise exc.OperationalError(f'Dictionary requires string keys, {key} has type {type(key)}')
            val = copy.deepcopy(val)
            if isinstance(val, dict):
                val = InlineDict(val)
            if isinstance(val, Expr):
                self.dict_items.append((key, len(self.components), None))
                self.components.append(val)
            else:
                self.dict_items.append((key, -1, val))

        self.type_spec: Optional[Dict[str, ColumnType]] = {}
        for key, idx, _ in self.dict_items:
            if idx == -1:
                # TODO: implement type inference for values
                self.type_spec = None
                break
            self.type_spec[key] = self.components[idx].col_type
        self.col_type = JsonType(self.type_spec)


    def _equals(self, other: 'InlineDict') -> bool:
        return self.dict_items == other.dict_items

    def sql_expr(self) -> Optional[sql.sql.expression.ClauseElement]:
        return None

    def eval(self, data_row: List[Any]) -> None:
        result = {}
        for key, idx, val in self.dict_items:
            assert isinstance(key, str)
            if idx >= 0:
                result[key] = data_row[self.components[idx].data_row_idx]
            else:
                result[key] = copy.deepcopy(val)
        data_row[self.data_row_idx] = result

    def _as_dict(self) -> Dict:
        return {'dict_items': self.dict_items, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr], t: catalog.Table) -> Expr:
        assert 'dict_items' in d
        arg: Dict[str, Any] = {}
        for key, idx, val in d['dict_items']:
            if idx >= 0:
                arg[key] = components[idx]
            else:
                arg[key] = val
        return cls(arg)


class InlineArray(Expr):
    """
    Array 'literal' which can use Exprs as values.
    """
    def __init__(self, elements: Tuple):
        # we need to call this in order to populate self.components
        super().__init__(ArrayType((len(elements),), ColumnType.Type.INT))

        # elements contains
        # - for Expr elements: (index into components, None)
        # - for non-Expr elements: (-1, value)
        self.elements: List[Tuple[int, Any]] = []
        for el in elements:
            el = copy.deepcopy(el)
            if isinstance(el, list):
                el = InlineArray(tuple(el))
            if isinstance(el, Expr):
                self.elements.append((len(self.components), None))
                self.components.append(el)
            else:
                self.elements.append((-1, el))

        element_type = InvalidType()
        for idx, val in self.elements:
            if idx >= 0:
                element_type = ColumnType.supertype(element_type, self.components[idx].col_type)
            else:
                element_type = ColumnType.supertype(element_type, ColumnType.get_value_type(val))
            if element_type is None:
                # there is no common element type: this is a json value, not an array
                # TODO: make sure this doesn't contain Images
                self.col_type = JsonType()
                return

        if element_type.is_scalar_type():
            self.col_type = ArrayType((len(self.elements),), element_type.type_enum)
        elif element_type.is_array_type():
            assert isinstance(element_type, ArrayType)
            self.col_type = ArrayType((len(self.elements), *element_type.shape), element_type.dtype)
        elif element_type.is_json_type():
            self.col_type = JsonType()


    def _equals(self, other: 'InlineDict') -> bool:
        return self.elements == other.elements

    def sql_expr(self) -> Optional[sql.sql.expression.ClauseElement]:
        return None

    def eval(self, data_row: List[Any]) -> None:
        result = [None] * len(self.elements)
        for i, (child_idx, val) in enumerate(self.elements):
            if child_idx >= 0:
                result[i] = data_row[self.components[child_idx].data_row_idx]
            else:
                result[i] = copy.deepcopy(val)
        data_row[self.data_row_idx] = np.array(result)

    def _as_dict(self) -> Dict:
        return {'elements': self.elements, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr], t: catalog.Table) -> Expr:
        assert 'elements' in d
        arg: List[Any] = []
        for idx, val in d['elements']:
            if idx >= 0:
                arg.append(components[idx])
            else:
                arg.append(val)
        return cls(tuple(arg))


class ArraySlice(Expr):
    """
    Slice operation on an array, eg, t.array_col[:, 1:2].
    """
    def __init__(self, arr: Expr, index: Tuple):
        assert arr.col_type.is_array_type()
        # determine result type
        super().__init__(arr.col_type)
        self.components = [arr]
        self.index = index

    @property
    def _array(self) -> Expr:
        return self.components[0]

    def _equals(self, other: 'ArraySlice') -> bool:
        return self.index == other.index

    def sql_expr(self) -> Optional[sql.sql.expression.ClauseElement]:
        return None

    def eval(self, data_row: List[Any]) -> None:
        val = data_row[self._array.data_row_idx]
        data_row[self.data_row_idx] = val[self.index]

    def _as_dict(self) -> Dict:
        index = []
        for el in self.index:
            if isinstance(el, slice):
                index.append([el.start, el.stop, el.step])
            else:
                index.append(el)
        return {'index': index, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr], t: catalog.Table) -> Expr:
        assert 'index' in d
        index = []
        for el in d['index']:
            if isinstance(el, list):
                index.append(slice(el[0], el[1], el[2]))
            else:
                index.append(el)
        return cls(components[0], tuple(index))


class Predicate(Expr):
    def __init__(self) -> None:
        super().__init__(BoolType())

    def extract_sql_predicate(self) -> Tuple[Optional[sql.sql.expression.ClauseElement], Optional['Predicate']]:
        """
        Return ClauseElement for what can be evaluated in SQL and a predicate for the remainder that needs to be
        evaluated in Python.
        Needed to for predicate push-down into SQL.
        """
        e = self.sql_expr()
        return (None, self) if e is None else (e, None)

    def split_conjuncts(
            self, condition: Callable[['Predicate'], bool]) -> Tuple[List['Predicate'], Optional['Predicate']]:
        """
        Returns clauses of a conjunction that meet condition in the first element.
        The second element contains remaining clauses, rolled into a conjunction.
        """
        if condition(self):
            return [self], None
        else:
            return [], self

    def __and__(self, other: object) -> 'CompoundPredicate':
        if not isinstance(other, Expr):
            raise TypeError(f'Other needs to be an expression: {type(other)}')
        if not other.col_type.is_bool_type():
            raise TypeError(f'Other needs to be an expression that returns a boolean: {other.col_type}')
        return CompoundPredicate(LogicalOperator.AND, [self, other])

    def __or__(self, other: object) -> 'CompoundPredicate':
        if not isinstance(other, Expr):
            raise TypeError(f'Other needs to be an expression: {type(other)}')
        if not other.col_type.is_bool_type():
            raise TypeError(f'Other needs to be an expression that returns a boolean: {other.col_type}')
        return CompoundPredicate(LogicalOperator.OR, [self, other])

    def __invert__(self) -> 'CompoundPredicate':
        return CompoundPredicate(LogicalOperator.NOT, [self])


class CompoundPredicate(Predicate):
    def __init__(self, operator: LogicalOperator, operands: List[Predicate]):
        super().__init__()
        self.operator = operator
        # operands are stored in self.components
        if self.operator == LogicalOperator.NOT:
            assert len(operands) == 1
            self.components = operands
        else:
            assert len(operands) > 1
            self.operands: List[Predicate] = []
            for operand in operands:
                self._merge_operand(operand)

    @classmethod
    def make_conjunction(cls, operands: List[Predicate]) -> Optional[Predicate]:
        if len(operands) == 0:
            return None
        if len(operands) == 1:
            return operands[0]
        return CompoundPredicate(LogicalOperator.AND, operands)

    def _merge_operand(self, operand: Predicate) -> None:
        """
        Merge this operand, if possible, otherwise simply record it.
        """
        if isinstance(operand, CompoundPredicate) and operand.operator == self.operator:
            # this can be merged
            for child_op in operand.components:
                self._merge_operand(child_op)
        else:
            self.components.append(operand)

    def _equals(self, other: 'CompoundPredicate') -> bool:
        return self.operator == other.operator

    def extract_sql_predicate(self) -> Tuple[Optional[sql.sql.expression.ClauseElement], Optional[Predicate]]:
        if self.operator == LogicalOperator.NOT:
            e = self.components[0].sql_expr()
            return (None, self) if e is None else (e, None)

        sql_exprs = [op.sql_expr() for op in self.components]
        if self.operator == LogicalOperator.OR and any(e is None for e in sql_exprs):
            # if any clause of a | can't be evaluated in SQL, we need to evaluate everything in Python
            return None, self
        if not(any(e is None for e in sql_exprs)):
            # we can do everything in SQL
            return self.sql_expr(), None

        assert self.operator == LogicalOperator.AND
        if not any(e is not None for e in sql_exprs):
            # there's nothing that can be done in SQL
            return None, self

        sql_preds = [e for e in sql_exprs if e is not None]
        other_preds = [self.components[i] for i, e in enumerate(sql_exprs) if e is None]
        assert len(sql_preds) > 0
        combined_sql_pred = sql.and_(*sql_preds)
        combined_other = self.make_conjunction(other_preds)
        return combined_sql_pred, combined_other

    def split_conjuncts(
            self, condition: Callable[['Predicate'], bool]) -> Tuple[List['Predicate'], Optional['Predicate']]:
        if self.operator == LogicalOperator.OR or self.operator == LogicalOperator.NOT:
            return super().split_conjuncts(condition)
        matches = [op for op in self.components if condition(op)]
        non_matches = [op for op in self.components if not condition(op)]
        return (matches, self.make_conjunction(non_matches))

    def sql_expr(self) -> Optional[sql.sql.expression.ClauseElement]:
        sql_exprs = [op.sql_expr() for op in self.components]
        if any(e is None for e in sql_exprs):
            return None
        if self.operator == LogicalOperator.NOT:
            assert len(sql_exprs) == 1
            return sql.not_(sql_exprs[0])
        assert len(sql_exprs) > 1
        operator = sql.and_ if self.operator == LogicalOperator.AND else sql.or_
        combined = operator(*sql_exprs)
        return combined

    def eval(self, data_row: List[Any]) -> None:
        if self.operator == LogicalOperator.NOT:
            data_row[self.data_row_idx] = not data_row[self.components[0].data_row_idx]
        else:
            val = True if self.operator == LogicalOperator.AND else False
            op_function = operator.and_ if self.operator == LogicalOperator.AND else operator.or_
            for op in self.components:
                val = op_function(val, data_row[op.data_row_idx])
            data_row[self.data_row_idx] = val

    def _as_dict(self) -> Dict:
        return {'operator': self.operator.value, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr], t: catalog.Table) -> Expr:
        assert 'operator' in d
        return cls(LogicalOperator(d['operator']), components)


class Comparison(Predicate):
    def __init__(self, operator: ComparisonOperator, op1: Expr, op2: Expr):
        super().__init__()
        self.operator = operator
        self.components = [op1, op2]

    def _equals(self, other: 'Comparison') -> bool:
        return self.operator == other.operator

    @property
    def _op1(self) -> Expr:
        return self.components[0]

    @property
    def _op2(self) -> Expr:
        return self.components[1]

    def sql_expr(self) -> Optional[sql.sql.expression.ClauseElement]:
        left = self._op1.sql_expr()
        right = self._op2.sql_expr()
        if left is None or right is None:
            return None
        if self.operator == ComparisonOperator.LT:
            return left < right
        if self.operator == ComparisonOperator.LE:
            return left <= right
        if self.operator == ComparisonOperator.EQ:
            return left == right
        if self.operator == ComparisonOperator.NE:
            return left != right
        if self.operator == ComparisonOperator.GT:
            return left > right
        if self.operator == ComparisonOperator.GE:
            return left >= right

    def eval(self, data_row: List[Any]) -> None:
        if self.operator == ComparisonOperator.LT:
            data_row[self.data_row_idx] = data_row[self._op1.data_row_idx] < data_row[self._op2.data_row_idx]
        elif self.operator == ComparisonOperator.LE:
            data_row[self.data_row_idx] = data_row[self._op1.data_row_idx] <= data_row[self._op2.data_row_idx]
        elif self.operator == ComparisonOperator.EQ:
            data_row[self.data_row_idx] = data_row[self._op1.data_row_idx] == data_row[self._op2.data_row_idx]
        elif self.operator == ComparisonOperator.NE:
            data_row[self.data_row_idx] = data_row[self._op1.data_row_idx] != data_row[self._op2.data_row_idx]
        elif self.operator == ComparisonOperator.GT:
            data_row[self.data_row_idx] = data_row[self._op1.data_row_idx] > data_row[self._op2.data_row_idx]
        elif self.operator == ComparisonOperator.GE:
            data_row[self.data_row_idx] = data_row[self._op1.data_row_idx] >= data_row[self._op2.data_row_idx]

    def _as_dict(self) -> Dict:
        return {'operator': self.operator.value, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr], t: catalog.Table) -> Expr:
        assert 'operator' in d
        return cls(ComparisonOperator(d['operator']), components[0], components[1])


class ImageSimilarityPredicate(Predicate):
    def __init__(self, img_col: ColumnRef, img: Optional[PIL.Image.Image] = None, text: Optional[str] = None):
        assert (img is None) != (text is None)
        super().__init__()
        self.img_col_ref = img_col
        self.components = [img_col]
        self.img = img
        self.text = text

    def embedding(self) -> np.ndarray:
        if self.text is not None:
            return clip.encode_text(self.text)
        else:
            return clip.encode_image(self.img)

    def _equals(self, other: 'ImageSimilarityPredicate') -> bool:
        return False

    def sql_expr(self) -> Optional[sql.sql.expression.ClauseElement]:
        return None

    def eval(self, data_row: List[Any]) -> None:
        assert False

    def _as_dict(self) -> Dict:
        assert False, 'not implemented'
        # TODO: convert self.img into a serializable string
        return {'img': self.img, 'text': self.text, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr], t: catalog.Table) -> Expr:
        assert 'img' in d
        assert 'text' in d
        assert len(components) == 1
        return cls(components[0], d['img'], d['text'])


class ArithmeticExpr(Expr):
    """
    Allows arithmetic exprs on json paths
    """
    def __init__(self, operator: ArithmeticOperator, op1: Expr, op2: Expr):
        if not op1.col_type.is_numeric_type() and not op1.col_type.is_json_type():
            raise exc.OperationalError(f'{operator} requires numeric type: {op1} has type {op1.col_type}')
        if not op2.col_type.is_numeric_type() and not op2.col_type.is_json_type():
            raise exc.OperationalError(f'{operator} requires numeric type: {op2} has type {op2.col_type}')
        # TODO: determine most specific common supertype
        if op1.col_type.is_json_type() or op2.col_type.is_json_type():
            # we assume it's a float
            super().__init__(FloatType())
        else:
            super().__init__(ColumnType.supertype(op1.col_type, op2.col_type))
        self.operator = operator
        self.components = [op1, op2]

    def _equals(self, other: 'ArithmeticExpr') -> bool:
        return self.operator == other.operator

    @property
    def _op1(self) -> Expr:
        return self.components[0]

    @property
    def _op2(self) -> Expr:
        return self.components[1]

    def sql_expr(self) -> Optional[sql.sql.expression.ClauseElement]:
        left = self._op1.sql_expr()
        right = self._op2.sql_expr()
        if left is None or right is None:
            return None
        if self.operator == ArithmeticOperator.ADD:
            return left + right
        if self.operator == ArithmeticOperator.SUB:
            return left - right
        if self.operator == ArithmeticOperator.MUL:
            return left * right
        if self.operator == ArithmeticOperator.DIV:
            return left / right
        if self.operator == ArithmeticOperator.MOD:
            return left % right

    def eval(self, data_row: List[Any]) -> None:
        op1_val = data_row[self._op1.data_row_idx]
        op2_val = data_row[self._op2.data_row_idx]
        # check types if we couldn't do that prior to execution
        if self._op1.col_type.is_json_type() and not isinstance(op1_val, int) and not isinstance(op1_val, float):
            raise exc.OperationalError(f'{self.operator} requires numeric type: {self._op1} has type {type(op1_val)}')
        if self._op2.col_type.is_json_type() and not isinstance(op2_val, int) and not isinstance(op2_val, float):
            raise exc.OperationalError(f'{self.operator} requires numeric type: {self._op2} has type {type(op2_val)}')

        if self.operator == ArithmeticOperator.ADD:
            data_row[self.data_row_idx] = op1_val + op2_val
        elif self.operator == ArithmeticOperator.SUB:
            data_row[self.data_row_idx] = op1_val - op2_val
        elif self.operator == ArithmeticOperator.MUL:
            data_row[self.data_row_idx] = op1_val * op2_val
        elif self.operator == ArithmeticOperator.DIV:
            data_row[self.data_row_idx] = op1_val / op2_val
        elif self.operator == ArithmeticOperator.MOD:
            data_row[self.data_row_idx] = op1_val % op2_val

    def _as_dict(self) -> Dict:
        return {'operator': self.operator.value, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr], t: catalog.Table) -> Expr:
        assert 'operator' in d
        assert len(components) == 2
        return cls(ArithmeticOperator(d['operator']), components[0], components[1])


class ObjectRef(Expr):
    """
    Reference to an intermediate result, such as the "scope variable" produced by a JsonMapper.
    The object is generated/materialized elsewhere and establishes a new scope.
    """
    def __init__(self, scope: ExprScope, owner: 'JsonMapper'):
        # TODO: do we need an Unknown type after all?
        super().__init__(JsonType())  # JsonType: this could be anything
        self._scope = scope
        self.owner = owner

    def scope(self) -> ExprScope:
        return self._scope

    def _equals(self, other: 'ObjectRef') -> bool:
        return self.owner is other.owner

    def sql_expr(self) -> Optional[sql.sql.expression.ClauseElement]:
        return None

    def eval(self, data_row: List[Any]) -> None:
        # this will be called, but the value has already been materialized elsewhere
        pass


class JsonMapper(Expr):
    """
    JsonMapper transforms the list output of a JsonPath by applying a target expr to every element of the list.
    The target expr would typically contain relative JsonPaths, which are bound to an ObjectRef, which in turn
    is populated by JsonMapper.eval(). The JsonMapper effectively creates a new scope for its target expr.
    """
    def __init__(self, src_expr: Expr, target_expr: Expr):
        # TODO: type spec should be List[target_expr.col_type]
        super().__init__(JsonType())

        # we're creating a new scope, but we don't know yet whether this is nested within another JsonMapper;
        # this gets resolved in bind_rel_paths(); for now we assume we're in the global scope
        self.target_expr_scope = ExprScope(_GLOBAL_SCOPE)

        scope_anchor = ObjectRef(self.target_expr_scope, self)
        self.components = [src_expr, target_expr, scope_anchor]
        self.parent_mapper: Optional[JsonMapper] = None
        self.evaluator: Optional[ExprEvaluator] = None

    def bind_rel_paths(self, mapper: Optional['JsonMapper']) -> None:
        self._src_expr.bind_rel_paths(mapper)
        self._target_expr.bind_rel_paths(self)
        self.parent_mapper = mapper
        parent_scope = _GLOBAL_SCOPE if mapper is None else mapper.target_expr_scope
        self.target_expr_scope.parent = parent_scope

    def scope(self) -> ExprScope:
        # need to ignore target_expr
        return self._src_expr.scope()

    def dependencies(self) -> List['Expr']:
        result = [self._src_expr]
        result.extend(self._target_dependencies(self._target_expr))
        return result

    def _target_dependencies(self, e: Expr) -> List['Expr']:
        """
        Return all subexprs of e of which the scope isn't contained in target_expr_scope.
        Those need to be evaluated before us.
        """
        expr_scope = e.scope()
        if not expr_scope.is_contained_in(self.target_expr_scope):
            return [e]
        result: List[Expr] = []
        for c in e.components:
            result.extend(self._target_dependencies(c))
        return result

    def equals(self, other: 'Expr') -> bool:
        """
        We override equals() because we need to avoid comparing our scope anchor.
        """
        if type(self) != type(other):
            return False
        return self._src_expr.equals(other._src_expr) and self._target_expr.equals(other._target_expr)

    @property
    def _src_expr(self) -> Expr:
        return self.components[0]

    @property
    def _target_expr(self) -> Expr:
        return self.components[1]

    @property
    def scope_anchor(self) -> Expr:
        return self.components[2]

    def _equals(self, other: 'JsonMapper') -> bool:
        return True

    def sql_expr(self) -> Optional[sql.sql.expression.ClauseElement]:
        return None

    def eval(self, data_row: List[Any]) -> None:
        # this will be called, but the value has already been materialized elsewhere
        src = data_row[self._src_expr.data_row_idx]
        if not isinstance(src, list):
            # invalid/non-list src path
            data_row[self.data_row_idx] = None
            return

        result = [None] * len(src)
        if self.evaluator is None:
            self.evaluator = ExprEvaluator([self._target_expr], None)
        for i, val in enumerate(src):
            data_row[self.scope_anchor.data_row_idx] = val
            # materialize target_expr
            self.evaluator.eval((), data_row)
            result[i] = data_row[self._target_expr.data_row_idx]
        data_row[self.data_row_idx] = result

    def _as_dict(self) -> Dict:
        """
        We need to avoid serializing component[2], which is an ObjectRef.
        """
        return {'components': [c.as_dict() for c in self.components[0:2]]}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr], t: catalog.Table) -> Expr:
        assert len(components) == 2
        return cls(components[0], components[1])


class ExprEvaluator:
    """
    Materializes values for a list of output exprs, subject to passing a filter.

    ex.: the select list [<img col 1>.alpha_composite(<img col 2>), <text col 3>]
    - sql row composition: [<file path col 1>, <file path col 2>, <text col 3>]
    - data row composition: [Image, str, Image, Image]
    - copy_exprs: [
        ColumnRef(data_row_idx: 2, sql_row_idx: 0, col: <col 1>)
        ColumnRef(data_row_idx: 3, sql_row_idx: 1, col: <col 2>)
        ColumnRef(data_row_idx: 1, sql_row_idx: 2, col: <col 3>)
      ]
    - eval_exprs: [ImageMethodCall(data_row_idx: 0, sql_row_id: -1)]
    """

    def __init__(self, output_exprs: List[Expr], filter: Optional[Predicate], with_sql: bool = True):
        # TODO: add self.literal_exprs so that we don't need to retrieve those from SQL
        # exprs that are materialized directly via SQL query and for which results can be copied from sql row
        # into data row
        self.filter_copy_exprs: List[Expr] = []
        self.output_copy_exprs: List[Expr] = []
        # exprs for which we need to call eval() to compute the value; must be called in the order stored here
        self.filter_eval_exprs: List[Expr] = []
        self.output_eval_exprs: List[Expr] = []
        self.filter = filter

        unique_ids: Set[int] = set()
        # analyze filter first, so that it can be evaluated before output_exprs
        if filter is not None:
            self._analyze_expr(filter, filter.scope(), self.filter_copy_exprs, self.filter_eval_exprs, unique_ids)
        for expr in output_exprs:
            self._analyze_expr(expr, expr.scope(), self.output_copy_exprs, self.output_eval_exprs, unique_ids)

    def _analyze_expr(
            self, expr: Expr, scope: ExprScope, copy_exprs: List[Expr], eval_exprs: List[Expr], unique_ids: Set[int]
    ) -> None:
        """
        Determine unique dependencies of expr and accumulate those in copy_exprs and eval_exprs.
        Dependencies that are not in 'scope' are assumed to have been materialized already and are ignored.
        """
        if expr.data_row_idx in unique_ids:
            return
        unique_ids.add(expr.data_row_idx)

        if expr.sql_row_idx >= 0:
            # this can be copied, no need to look at its dependencies
            copy_exprs.append(expr)
            return

        for d in expr.dependencies():
            if d.scope() != scope:
                continue
            self._analyze_expr(d, scope, copy_exprs, eval_exprs, unique_ids)
        # make sure to eval() this after its dependencies
        eval_exprs.append(expr)

    def eval(self, sql_row: Tuple[Any], data_row: List[Any]) -> bool:
        """
        If the filter predicate evaluates to True, populates the data_row slots of the output_exprs.
        """
        if self.filter is not None:
            # we need to evaluate the remaining filter predicate first
            self._copy_to_data_row(self.filter_copy_exprs, sql_row, data_row)
            for expr in self.filter_eval_exprs:
                expr.eval(data_row)
            if not data_row[self.filter.data_row_idx]:
                return False

        # materialize output_exprs
        self._copy_to_data_row(self.output_copy_exprs, sql_row, data_row)
        for expr in self.output_eval_exprs:
            expr.eval(data_row)
        return True

    def _copy_to_data_row(self, exprs: List[Expr], sql_row: Tuple[Any], data_row: List[Any]):
        """
        Copy expr values from sql to data row.
        """
        for expr in exprs:
            assert expr.sql_row_idx != -1
            if expr.col_type.is_image_type():
                # column value is a file path that we need to open
                file_path = sql_row[expr.sql_row_idx]
                try:
                    img = PIL.Image.open(file_path)
                    #img.thumbnail((128, 128))
                    data_row[expr.data_row_idx] = img
                except Exception:
                    raise exc.OperationalError(f'Error reading image file: {file_path}')
            elif expr.col_type.is_array_type():
                # column value is a saved numpy array
                array_data = sql_row[expr.sql_row_idx]
                data_row[expr.data_row_idx] = np.load(io.BytesIO(array_data))
            else:
                data_row[expr.data_row_idx] = sql_row[expr.sql_row_idx]


class UniqueExprSet:
    """
    We want to avoid duplicate expr evaluation, so we keep track of unique exprs (duplicates share the
    same data_row_idx). However, __eq__() doesn't work for sets, so we use a list here.
    """
    def __init__(self):
        self.unique_exprs: List[Expr] = []

    def add(self, expr: Expr) -> bool:
        """
        If expr is not unique, sets expr.data/sql_row_idx to that of the already-recorded duplicate and returns
        False, otherwise returns True.
        """
        try:
            existing = next(e for e in self.unique_exprs if e.equals(expr))
            expr.data_row_idx = existing.data_row_idx
            expr.sql_row_idx = existing.sql_row_idx
            return False
        except StopIteration:
            self.unique_exprs.append(expr)
            return True

    def __iter__(self) -> Iterator[Expr]:
        return iter(self.unique_exprs)


class ExprEvalCtx:
    """
    Assigns execution state necessary to materialize a list of Exprs into a data row:
    - Expr.sql_/data_row_idx

    Data row:
    - List[Any]
    - contains slots for all materialized component exprs (ie, not for predicates that turn into the SQL Where clause):
    a) every DataFrame.select_list expr
    b) the parts of the where clause predicate that cannot be evaluated in SQL
    b) every component expr of a) and b), recursively
    - IMAGE columns are materialized immediately as a PIL.Image.Image

    ex.: the select list [<img col 1>.alpha_composite(<img col 2>), <text col 3>]
    - sql row composition: [<file path col 1>, <file path col 2>, <text col 3>]
    - data row composition: [Image, str, Image, Image]
    """

    def __init__(self, output_exprs: List[Expr], filter: Optional[Predicate]):
        """
        Init for list of materialized exprs and a possible filter.
        with_sql == True: if an expr e has a e.sql_expr(), its components do not need to be materialized
        (and consequently also don't get data_row_idx assigned) and the expr value is produced via a Select stmt
        """

        # objects needed to materialize the SQL result row
        self.sql_exprs: List[sql.sql.expression.ClauseElement] = []
        self.unique_exprs = UniqueExprSet()
        self.next_data_row_idx = 0

        if filter is not None:
            self._analyze_expr(filter)
        for expr in output_exprs:
            self._analyze_expr(expr)

    @property
    def num_materialized(self) -> int:
        return self.next_data_row_idx

    def _analyze_expr(self, expr: Expr) -> None:
        """
        Assign Expr.data_row_idx and Expr.sql_row_idx.
        """
        if not self.unique_exprs.add(expr):
            # nothing left to do
            return

        sql_expr = expr.sql_expr()
        # if this can be materialized via SQL we don't need to look at its components;
        # we special-case Literals because we don't want to have to materialize them via SQL
        if sql_expr is not None and not isinstance(expr, Literal):
            assert expr.data_row_idx < 0
            expr.data_row_idx = self.next_data_row_idx
            self.next_data_row_idx += 1
            expr.sql_row_idx = len(self.sql_exprs)
            self.sql_exprs.append(sql_expr)
            return

        # expr value needs to be computed via Expr.eval()
        for c in expr.components:
            self._analyze_expr(c)
        assert expr.data_row_idx < 0
        expr.data_row_idx = self.next_data_row_idx
        self.next_data_row_idx += 1


class ComputedColEvalCtx:
    """
    EvalCtx for computed cols:
    - referenced inputs are not supplied via SQL
    - a col's ColumnRef and value_expr need to share the same data_row_idx
    """

    def __init__(self, computed_col_info: List[Tuple[ColumnRef, Expr]]):
        """
        computed_col_info: list of (ref to col, value_expr of col)
        """

        # we want to avoid duplicate expr evaluation, so we keep track of unique exprs (duplicates share the
        # same data_row_idx); however, __eq__() doesn't work for sets, so we use a list here
        self.unique_exprs = UniqueExprSet()
        self.next_data_row_idx = 0

        for col_ref, expr in computed_col_info:
            self._analyze_expr(expr)
            # the expr materializes the value of that column
            col_ref.data_row_idx = expr.data_row_idx
            # future references to that column will use the already-assigned data_row_idx
            self.unique_exprs.add(col_ref)

    @property
    def num_materialized(self) -> int:
        return self.next_data_row_idx

    def _analyze_expr(self, expr: Expr) -> None:
        """
        Assign Expr.data_row_idx.
        """
        if not self.unique_exprs.add(expr):
            # nothing left to do
            return
        for c in expr.components:
            self._analyze_expr(c)
        assert expr.data_row_idx < 0
        expr.data_row_idx = self.next_data_row_idx
        self.next_data_row_idx += 1



from typing import Optional
from pathlib import Path
import glob

from pixeltable.env import Env


def is_computed_img_path(path: str) -> bool:
    try:
        _ = Path(path).relative_to(Env.get().img_dir)
        return True
    except ValueError:
        return False

def get_computed_img_path(tbl_id: int, col_id: int, version: int, rowid: int) -> str:
    return Env.get().img_dir / f'img_{tbl_id}_{col_id}_{version}_{rowid}.jpg'

def get_extracted_frame_path(tbl_id: int, video_col_id: int, version: int, offset: int) -> str:
    return Env.get().img_dir / f'frame_{tbl_id}_{video_col_id}_{version}_{offset}'

def computed_imgs(
        tbl_id: Optional[int] = None, col_id: Optional[int] = None, version: Optional[int] = None) -> int:
    path = f'{Env.get().img_dir}/img_'
    path += f'{tbl_id}_' if tbl_id is not None else '*_'
    path += f'{col_id}_' if col_id is not None else '*_'
    path += f'{version}_' if version is not None else '*_'
    path += '*'
    names = glob.glob(path)
    return names

def computed_img_count(
        tbl_id: Optional[int] = None, col_id: Optional[int] = None, version: Optional[int] = None) -> int:
    return len(computed_imgs(tbl_id=tbl_id, col_id=col_id, version=version))

def extracted_frames(tbl_id: Optional[int] = None, version: Optional[int] = None) -> int:
    path = f'{Env.get().img_dir}/frame_'
    path += f'{tbl_id}_' if tbl_id is not None else '*_'
    path += '*_'  # video_col_id
    path += f'{version}_' if version is not None else '*_'
    path += '*_'  # offset
    path += '*'  # running frame index
    names = glob.glob(path)
    return names

def extracted_frame_count(tbl_id: Optional[int] = None, version: Optional[int] = None) -> int:
    return len(extracted_frames(tbl_id, version))



import enum
import platform

import sqlalchemy as sql
from sqlalchemy import Integer, String, Enum, Boolean, TIMESTAMP, BigInteger, LargeBinary, JSON
from sqlalchemy import ForeignKey, UniqueConstraint, ForeignKeyConstraint
from sqlalchemy.orm import declarative_base

from pixeltable import type_system as pt_types

Base = declarative_base()


class Db(Base):
    __tablename__ = 'dbs'

    id = sql.Column(Integer, primary_key=True, autoincrement=True, nullable=False)
    name = sql.Column(String, nullable=False)


class Dir(Base):
    __tablename__ = 'dirs'

    id = sql.Column(Integer, primary_key=True, autoincrement=True, nullable=False)
    db_id = sql.Column(Integer, ForeignKey('dbs.id'), nullable=False)
    path = sql.Column(String, nullable=False)
    is_snapshot = sql.Column(Boolean, nullable=False)


class Table(Base):
    __tablename__ = 'tables'

    MAX_VERSION = 9223372036854775807  # 2^63 - 1

    id = sql.Column(Integer, primary_key=True, autoincrement=True, nullable=False)
    db_id = sql.Column(Integer, ForeignKey('dbs.id'), nullable=False)
    dir_id = sql.Column(Integer, ForeignKey('dirs.id'), nullable=False)
    name = sql.Column(String, nullable=False)
    parameters = sql.Column(JSON, nullable=False)

    # monotonically increasing w/in Table for both data and schema changes, starting at 0
    current_version = sql.Column(BigInteger, nullable=False)
    # each version has a corresponding schema version (current_version >= current_schema_version)
    current_schema_version = sql.Column(BigInteger, nullable=False)

    # if False, can't apply schema or data changes to this table
    # (table got dropped, but we need to keep a record of it for snapshots)
    is_mutable = sql.Column(Boolean, nullable=False)

    next_col_id = sql.Column(Integer, nullable=False)  # used to assign Column.id

    # - used to assign the rowid column in the storage table
    # - every row is assigned a unique and immutable rowid on insertion
    next_row_id = sql.Column(BigInteger, nullable=False)

    __table_args__ = (
        #ForeignKeyConstraint(
            #['id', 'current_schema_version'], ['tableschemaversions.tbl_id', 'tableschemaversions.schema_version']),
    )

    def storage_name(self) -> str:
        return f'tbl_{self.db_id}_{self.id}'


# versioning: each table schema change results in a new record
class TableSchemaVersion(Base):
    __tablename__ = 'tableschemaversions'

    tbl_id = sql.Column(Integer, ForeignKey('tables.id'), primary_key=True, nullable=False)
    schema_version = sql.Column(BigInteger, primary_key=True, nullable=False)
    preceding_schema_version = sql.Column(BigInteger, nullable=False)


# - records the physical schema of a table, ie, the columns that are actually stored
# - versioning information is needed to GC unreachable storage columns
# - one record per column (across all schema versions)
class StorageColumn(Base):
    __tablename__ = 'storagecolumns'

    tbl_id = sql.Column(Integer, ForeignKey('tables.id'), primary_key=True, nullable=False)
    # immutable and monotonically increasing from 0 w/in Table
    col_id = sql.Column(Integer, primary_key=True, nullable=False)
    # table schema version when col was added
    schema_version_add = sql.Column(BigInteger, nullable=False)
    # table schema version when col was dropped
    schema_version_drop = sql.Column(BigInteger, nullable=True)

    __table_args__ = (
        ForeignKeyConstraint(
            ['tbl_id', 'schema_version_add'], ['tableschemaversions.tbl_id', 'tableschemaversions.schema_version']),
        ForeignKeyConstraint(
            ['tbl_id', 'schema_version_drop'], ['tableschemaversions.tbl_id', 'tableschemaversions.schema_version'])
    )


# - records the logical (user-visible) schema of a table
# - contains the full set of columns for each new schema version: one record per column x schema version
class SchemaColumn(Base):
    __tablename__ = 'schemacolumns'

    tbl_id = sql.Column(Integer, ForeignKey('tables.id'), primary_key=True, nullable=False)
    schema_version = sql.Column(BigInteger, primary_key=True, nullable=False)
    # immutable and monotonically increasing from 0 w/in Table
    col_id = sql.Column(Integer, primary_key=True, nullable=False)
    pos = sql.Column(Integer, nullable=False)  # position in table, starting at 0
    name = sql.Column(String, nullable=False)
    col_type = sql.Column(String, nullable=False)  # json
    is_nullable = sql.Column(Boolean, nullable=False)
    is_pk = sql.Column(Boolean, nullable=False)
    value_expr = sql.Column(String, nullable=True)  # json
    # if True, creates vector index for this column
    is_indexed = sql.Column(Boolean, nullable=False)

    __table_args__ = (
        UniqueConstraint('tbl_id', 'schema_version', 'pos'),
        UniqueConstraint('tbl_id', 'schema_version', 'name'),
        ForeignKeyConstraint(['tbl_id', 'col_id'], ['storagecolumns.tbl_id', 'storagecolumns.col_id']),
        ForeignKeyConstraint(
            ['tbl_id', 'schema_version'], ['tableschemaversions.tbl_id', 'tableschemaversions.schema_version'])
    )


class TableSnapshot(Base):
    __tablename__ = 'tablesnapshots'

    id = sql.Column(Integer, primary_key=True, autoincrement=True, nullable=False)
    db_id = sql.Column(Integer, ForeignKey('dbs.id'), nullable=False)
    dir_id = sql.Column(Integer, ForeignKey('dirs.id'), nullable=False)
    name = sql.Column(String, nullable=False)
    tbl_id = sql.Column(Integer, nullable=False)
    tbl_version = sql.Column(BigInteger, nullable=False)
    tbl_schema_version = sql.Column(BigInteger, nullable=False)

    __table_args__ = (
        ForeignKeyConstraint(['tbl_id'], ['tables.id']),
        ForeignKeyConstraint(
            ['tbl_id', 'tbl_schema_version'], ['tableschemaversions.tbl_id', 'tableschemaversions.schema_version']),
    )


class Function(Base):
    """
    User-defined functions that are not library functions (ie, aren't available at runtime as a symbol in a known
    module).
    Functions without a name are anonymous functions used in the definition of a computed column.
    Functions that have names are also assigned to a database and directory.
    We store the Python version under which a Function was created (and the callable pickled) in order to warn
    against version mismatches.
    """
    __tablename__ = 'functions'

    id = sql.Column(Integer, primary_key=True, autoincrement=True, nullable=False)
    db_id = sql.Column(Integer, ForeignKey('dbs.id'), nullable=True)
    dir_id = sql.Column(Integer, ForeignKey('dirs.id'), nullable=True)
    name = sql.Column(String, nullable=True)
    return_type = sql.Column(String, nullable=False)  # json
    param_types = sql.Column(String, nullable=False)  # json
    eval_obj = sql.Column(LargeBinary, nullable=True)  # Function.eval_fn
    init_obj = sql.Column(LargeBinary, nullable=True)  # AggregateFunction.init_fn
    update_obj = sql.Column(LargeBinary, nullable=True)  # AggregateFunction.update_fn
    value_obj = sql.Column(LargeBinary, nullable=True)  # AggregateFunction.value_fn
    python_version = sql.Column(
        String, nullable=False, default=platform.python_version, onupdate=platform.python_version)


class OpCodes(enum.Enum):
    CREATE_DB = 1
    RENAME_DB = 1
    DROP_DB = 1
    CREATE_TABLE = 1
    RENAME_TABLE = 1
    DROP_TABLE = 1
    ADD_COLUMN = 1
    RENAME_COLUMN = 1
    DROP_COLUMN = 1
    CREATE_DIR = 1
    DROP_DIR = 1
    CREATE_SNAPSHOT = 1
    DROP_SNAPSHOT = 1


class Operation(Base):
    __tablename__ = 'oplog'

    id = sql.Column(Integer, primary_key=True, autoincrement=True, nullable=False)
    ts = sql.Column(TIMESTAMP, nullable=False)
    # encodes db_id, schema object type (table, view, table/view snapshot), table/view/... id
    schema_object = sql.Column(BigInteger, nullable=False)
    opcode = sql.Column(Enum(OpCodes), nullable=False)
    # operation-specific details; json
    details = sql.Column(String, nullable=False)



import numpy as np
import pandas as pd
import pytest

from pixeltable.function import Function, FunctionRegistry
from pixeltable.type_system import IntType
from pixeltable import catalog
import pixeltable as pt
from pixeltable import exceptions as exc


def dummy_fn(i: int) -> int:
    return i

class TestFunction:
    eval_fn = lambda x: x + 1
    func = Function(IntType(), [IntType()], eval_fn=eval_fn)

    def test_serialize_anonymous(self, init_db: None) -> None:
        d = self.func.as_dict()
        FunctionRegistry.get().clear_cache()
        deserialized = Function.from_dict(d)
        assert deserialized.eval_fn(1) == 2

    def test_create(self, test_db: catalog.Db) -> None:
        db = test_db
        db.create_function('test_fn', self.func)
        FunctionRegistry.get().clear_cache()
        cl = pt.Client()
        db2 = cl.get_db('test')
        fn2 = db2.load_function('test_fn')
        assert fn2.eval_fn(1) == 2

        with pytest.raises(exc.DuplicateNameError):
            db.create_function('test_fn', self.func)
        with pytest.raises(exc.UnknownEntityError):
            db.create_function('dir1.test_fn', self.func)
        with pytest.raises(exc.Error):
            library_fn = Function(IntType(), [IntType()], module_name=__name__, eval_symbol='dummy_fn')
            db.create_function('library_fn', library_fn)

    def test_update(self, test_db: catalog.Db) -> None:
        db = test_db
        db.create_function('test_fn', self.func)
        FunctionRegistry.get().clear_cache()
        cl = pt.Client()
        db2 = cl.get_db('test')
        fn = db2.load_function('test_fn')
        db2.update_function('test_fn', lambda x: x + 2)
        # change visible in previously-loaded Function
        assert fn.eval_fn(1) == 3

    def test_rename(self, test_db: catalog.Db) -> None:
        db = test_db
        db.create_function('test_fn', self.func)

        FunctionRegistry.get().clear_cache()
        cl = pt.Client()
        db2 = cl.get_db('test')
        with pytest.raises(exc.UnknownEntityError):
            db2.rename_function('test_fn2', 'test_fn')
        db2.rename_function('test_fn', 'test_fn2')
        func = db2.load_function('test_fn2')
        assert func.eval_fn(1) == 2

        with pytest.raises(exc.UnknownEntityError):
            _ = db2.load_function('test_fn')

        # move function between directories
        db2.create_dir('functions')
        db2.create_dir('functions2')
        db2.create_function('functions.func1', self.func)
        with pytest.raises(exc.UnknownEntityError):
            db2.rename_function('functions2.func1', 'functions.func1')
        db2.rename_function('functions.func1', 'functions2.func1')

        FunctionRegistry.get().clear_cache()
        cl = pt.Client()
        db3 = cl.get_db('test')
        func = db3.load_function('functions2.func1')
        assert func.eval_fn(1) == 2
        with pytest.raises(exc.UnknownEntityError):
            _ = db3.load_function('functions.func1')

    def test_drop(self, test_db: catalog.Db) -> None:
        db = test_db
        db.create_function('test_fn', self.func)
        FunctionRegistry.get().clear_cache()
        cl = pt.Client()
        db2 = cl.get_db('test')
        db2.drop_function('test_fn')

        with pytest.raises(exc.UnknownEntityError):
            _ = db2.load_function('test_fn')



from typing import Any, Tuple
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
import tensorflow as tf

from  pixeltable.type_system import ColumnType, JsonType

TF_DTYPES = {
    # ColumnType.DType.INT8: tf.int8,
    # ColumnType.DType.INT16: tf.int16,
    # ColumnType.DType.INT32: tf.int32,
    # ColumnType.DType.INT64: tf.int64,
    # ColumnType.DType.UINT8: tf.uint8,
    # ColumnType.DType.UINT16: tf.uint16,
    # ColumnType.DType.UINT32: tf.uint32,
    # ColumnType.DType.UINT64: tf.uint64,
    ColumnType.Type.INT: tf.int64,
    ColumnType.Type.FLOAT: tf.float32
}

def to_tensor(v: Any, t: ColumnType) -> tf.Tensor:
    if t.is_scalar_type():
        return tf.convert_to_tensor(v)
    if t.is_image_type():
        # convert PIL.Image.Image to tf.Tensor
        return tf.cast(tf.keras.utils.img_to_array(v), tf.uint8)
    if t.is_array_type():
        return tf.convert_to_tensor(v, dtype=TF_DTYPES[t.dtype])
    if t.is_json_type():
        assert isinstance(v, dict)
        assert isinstance(t, JsonType)
        assert t.type_spec is not None
        return {key: to_tensor(val, t.type_spec[key]) for key, val in v.items()}



from pixeltable.type_system import StringType, ImageType, ArrayType, ColumnType
from pixeltable.function import Function


encode_image = Function(
    ArrayType((512,), ColumnType.Type.FLOAT), [ImageType()],
    module_name='pixeltable.utils.clip', eval_symbol='encode_image')
encode_text = Function(
    ArrayType((512,), ColumnType.Type.FLOAT), [StringType()],
    module_name = 'pixeltable.utils.clip', eval_symbol = 'encode_text')



import numpy as np
import PIL.Image
import clip
import torch
import PIL.Image


_device = 'cuda' if torch.cuda.is_available() else 'cpu'
_model, _preprocess = clip.load("ViT-B/32", device=_device)

def encode_image(img: PIL.Image.Image) -> np.ndarray:
    preprocessed = _preprocess(img).unsqueeze(0).to(_device)
    features = _model.encode_image(preprocessed)
    val = features.numpy(force=True).squeeze()
    return val

def encode_text(txt: str) -> np.ndarray:
    preprocessed = clip.tokenize([txt]).to(_device)
    features = _model.encode_text(preprocessed)
    val = features.numpy(force=True).squeeze()
    return val



from pixeltable.type_system import FloatType, ImageType
from pixeltable.function import Function


alpha_composite = Function(
    ImageType(), [ImageType(), ImageType()], module_name='PIL.Image', eval_symbol='alpha_composite')
blend = Function(ImageType(), [ImageType(), ImageType(), FloatType()], module_name='PIL.Image', eval_symbol='blend')
composite = Function(
    ImageType(), [ImageType(), ImageType(), ImageType()], module_name='PIL.Image', eval_symbol='composite')



import abc
from typing import Any, Optional, Tuple, Dict, Callable, List, Union
import enum
import datetime
import json

import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
#import tensorflow as tf
import PIL.Image
import sqlalchemy as sql



class ColumnType:
    @enum.unique
    class Type(enum.Enum):
        STRING = 0
        INT = 1
        FLOAT = 2
        BOOL = 3
        TIMESTAMP = 4
        JSON = 5
        ARRAY = 6
        IMAGE = 7
        VIDEO = 8

        # exprs that don't evaluate to a computable value in Pixeltable, such as an Image member function
        INVALID = 9

        def to_tf(self) -> 'tf.dtypes.DType':
            import tensorflow as tf
            if self == self.STRING:
                return tf.string
            if self == self.INT:
                return tf.int64
            if self == self.FLOAT:
                return tf.float32
            if self == self.BOOL:
                return tf.bool
            raise TypeError(f'Cannot convert {self} to TensorFlow')

        @classmethod
        def supertype(
                cls, type1: 'Type', type2: 'Type',
                # we need to pass this in because we can't easily add it as a class member
                common_supertypes: Dict[Tuple['Type', 'Type'], 'Type']
        ) -> Optional['Type']:
            if type1 == type2:
                return type1
            t = common_supertypes.get((type1, type2))
            if t is not None:
                return t
            t = common_supertypes.get((type2, type1))
            if t is not None:
                return t
            return None


    @enum.unique
    class DType(enum.Enum):
        """
        Base type used in images and arrays
        """
        BOOL = 0,
        INT8 = 1,
        INT16 = 2,
        INT32 = 3,
        INT64 = 4,
        UINT8 = 5,
        UINT16 = 6,
        UINT32 = 7,
        UINT64 = 8,
        FLOAT16 = 9,
        FLOAT32 = 10,
        FLOAT64 = 11

    scalar_types = {Type.STRING, Type.INT, Type.FLOAT, Type.BOOL, Type.TIMESTAMP}
    numeric_types = {Type.INT, Type.FLOAT}
    common_supertypes: Dict[Tuple[Type, Type], Type] = {
        (Type.BOOL, Type.INT): Type.INT,
        (Type.BOOL, Type.FLOAT): Type.FLOAT,
        (Type.INT, Type.FLOAT): Type.FLOAT,
    }

    def __init__(self, t: Type):
        self._type = t

    @property
    def type_enum(self) -> Type:
        return self._type

    def serialize(self) -> str:
        return json.dumps(self.as_dict())

    @classmethod
    def serialize_list(cls, type_list: List['ColumnType']) -> str:
        return json.dumps([t.as_dict() for t in type_list])

    def as_dict(self) -> Dict:
        return {
            '_classname': self.__class__.__name__,
            **self._as_dict(),
        }

    def _as_dict(self) -> Dict:
        return {}

    @classmethod
    def deserialize(cls, type_str: str) -> 'ColumnType':
        type_dict = json.loads(type_str)
        return cls.from_dict(type_dict)

    @classmethod
    def deserialize_list(cls, type_list_str: str) -> List['ColumnType']:
        type_dict_list = json.loads(type_list_str)
        return [cls.from_dict(type_dict) for type_dict in type_dict_list]

    @classmethod
    def from_dict(cls, type_dict: Dict) -> 'ColumnType':
        assert '_classname' in type_dict
        type_class = globals()[type_dict['_classname']]
        return type_class._from_dict(type_dict)

    @classmethod
    def _from_dict(cls, d: Dict) -> 'ColumnType':
        """
        Default implementation: simply invoke c'tor without arguments
        """
        return cls()

    @classmethod
    def make_type(cls, t: Type) -> 'ColumnType':
        assert t != cls.Type.INVALID and t != cls.Type.ARRAY
        if t == cls.Type.STRING:
            return StringType()
        if t == cls.Type.INT:
            return IntType()
        if t == cls.Type.FLOAT:
            return FloatType()
        if t == cls.Type.BOOL:
            return BoolType()
        if t == cls.Type.TIMESTAMP:
            return TimestampType()
        if t == cls.Type.JSON:
            return JsonType()
        if t == cls.Type.IMAGE:
            return ImageType()
        if t == cls.Type.VIDEO:
            return VideoType()

    def __str__(self) -> str:
        return self._type.name.lower()

    def __eq__(self, other: object) -> bool:
        assert isinstance(other, ColumnType)
        if False and type(self) != type(other):
            return False
        for member_var in vars(self).keys():
            if getattr(self, member_var) != getattr(other, member_var):
                return False
        return True

    @classmethod
    def supertype(cls, type1: 'ColumnType', type2: 'ColumnType') -> Optional['ColumnType']:
        if type1 == type2:
            return type1

        if type1.is_invalid_type():
            return type2
        if type2.is_invalid_type():
            return type1

        if type1.is_scalar_type() and type2.is_scalar_type():
            t = cls.Type.supertype(type1._type, type2._type, cls.common_supertypes)
            if t is not None:
                return cls.make_type(t)
            return None

        if type1._type == type2._type:
            return cls._supertype(type1, type2)

        return None

    @classmethod
    @abc.abstractmethod
    def _supertype(cls, type1: 'ColumnType', type2: 'ColumnType') -> Optional['ColumnType']:
        """
        Class-specific implementation of determining the supertype. type1 and type2 are from the same subclass of
        ColumnType.
        """
        pass


    @classmethod
    def get_value_type(cls, val: Any) -> 'ColumnType':
        if isinstance(val, str):
            return StringType()
        if isinstance(val, int):
            return IntType()
        if isinstance(val, float):
            return FloatType()
        if isinstance(val, bool):
            return BoolType()
        if isinstance(val, datetime.datetime) or isinstance(val, datetime.date):
            return TimestampType()

    def is_scalar_type(self) -> bool:
        return self._type in self.scalar_types

    def is_numeric_type(self) -> bool:
        return self._type in self.numeric_types

    def is_invalid_type(self) -> bool:
        return self._type == self.Type.INVALID

    def is_string_type(self) -> bool:
        return self._type == self.Type.STRING

    def is_int_type(self) -> bool:
        return self._type == self.Type.INT

    def is_float_type(self) -> bool:
        return self._type == self.Type.FLOAT

    def is_bool_type(self) -> bool:
        return self._type == self.Type.BOOL

    def is_timestamp_type(self) -> bool:
        return self._type == self.Type.TIMESTAMP

    def is_json_type(self) -> bool:
        return self._type == self.Type.JSON

    def is_array_type(self) -> bool:
        return self._type == self.Type.ARRAY

    def is_image_type(self) -> bool:
        return self._type == self.Type.IMAGE

    def is_video_type(self) -> bool:
        return self._type == self.Type.VIDEO

    @abc.abstractmethod
    def to_sql(self) -> str:
        """
        Return corresponding Postgres type.
        """
        pass

    @abc.abstractmethod
    def to_sa_type(self) -> Any:
        """
        Return corresponding SQLAlchemy type.
        return type Any: there doesn't appear to be a superclass for the sqlalchemy types
        """
        assert self._type != self.Type.INVALID
        if self._type == self.Type.STRING:
            return sql.String
        if self._type == self.Type.INT:
            return sql.Integer
        if self._type == self.Type.FLOAT:
            return sql.Float
        if self._type == self.Type.BOOL:
            return sql.Boolean
        if self._type == self.Type.TIMESTAMP:
            return sql.TIMESTAMP
        if self._type == self.Type.IMAGE:
            # the URL
            return sql.String
        if self._type == self.Type.JSON:
            return sql.dialects.postgresql.JSONB
        if self._type == self.Type.ARRAY:
            return sql.VARBINARY
        assert False

    @staticmethod
    def no_conversion(v: Any) -> Any:
        """
        Special return value of conversion_fn() that indicates that no conversion is necessary.
        Should not be called
        """
        assert False

    def conversion_fn(self, target: 'ColumnType') -> Optional[Callable[[Any], Any]]:
        """
        Return Callable that converts a column value of type self to a value of type 'target'.
        Returns None if conversion isn't possible.
        """
        return None

    @abc.abstractmethod
    def to_tf(self) -> Union['tf.TypeSpec', Dict[str, 'tf.TypeSpec']]:
        pass


class InvalidType(ColumnType):
    def __init__(self):
        super().__init__(self.Type.INVALID)

    def to_sql(self) -> str:
        assert False

    def to_sa_type(self) -> Any:
        assert False

    def to_tf(self) -> Union['tf.TypeSpec', Dict[str, 'tf.TypeSpec']]:
        raise TypeError(f'Invalid type cannot be converted to Tensorflow')


class StringType(ColumnType):
    def __init__(self):
        super().__init__(self.Type.STRING)

    def conversion_fn(self, target: ColumnType) -> Optional[Callable[[Any], Any]]:
        if not target.is_timestamp_type():
            return None
        def convert(val: str) -> Optional[datetime]:
            try:
                dt = datetime.datetime.fromisoformat(val)
                return dt
            except ValueError:
                return None
        return convert

    def to_sql(self) -> str:
        return 'VARCHAR'

    def to_sa_type(self) -> str:
        return sql.String

    def to_tf(self) -> Union['tf.TypeSpec', Dict[str, 'tf.TypeSpec']]:
        import tensorflow as tf
        return tf.TensorSpec(shape=(), dtype=tf.string)


class IntType(ColumnType):
    def __init__(self):
        super().__init__(self.Type.INT)

    def to_sql(self) -> str:
        return 'INTEGER'

    def to_sa_type(self) -> str:
        return sql.Integer

    def to_tf(self) -> Union['tf.TypeSpec', Dict[str, 'tf.TypeSpec']]:
        # TODO: how to specify the correct int subtype?
        import tensorflow as tf
        return tf.TensorSpec(shape=(), dtype=tf.int64)


class FloatType(ColumnType):
    def __init__(self):
        super().__init__(self.Type.FLOAT)

    def to_sql(self) -> str:
        return 'FLOAT'

    def to_sa_type(self) -> str:
        return sql.Float

    def to_tf(self) -> Union['tf.TypeSpec', Dict[str, 'tf.TypeSpec']]:
        import tensorflow as tf
        # TODO: how to specify the correct float subtype?
        return tf.TensorSpec(shape=(), dtype=tf.float32)


class BoolType(ColumnType):
    def __init__(self):
        super().__init__(self.Type.BOOL)

    def to_sql(self) -> str:
        return 'BOOLEAN'

    def to_sa_type(self) -> str:
        return sql.Boolean

    def to_tf(self) -> Union['tf.TypeSpec', Dict[str, 'tf.TypeSpec']]:
        import tensorflow as tf
        # TODO: how to specify the correct int subtype?
        return tf.TensorSpec(shape=(), dtype=tf.bool)


class TimestampType(ColumnType):
    def __init__(self):
        super().__init__(self.Type.TIMESTAMP)

    def to_sql(self) -> str:
        return 'INTEGER'

    def to_sa_type(self) -> str:
        return sql.TIMESTAMP

    def to_tf(self) -> Union['tf.TypeSpec', Dict[str, 'tf.TypeSpec']]:
        raise TypeError(f'Timestamp type cannot be converted to Tensorflow')


class JsonType(ColumnType):
    # TODO: type_spec also needs to be able to express lists
    def __init__(self, type_spec: Optional[Dict[str, ColumnType]] = None):
        super().__init__(self.Type.JSON)
        self.type_spec = type_spec

    def _as_dict(self) -> Dict:
        result = super()._as_dict()
        if self.type_spec is not None:
            type_spec_dict = {field_name: field_type.serialize() for field_name, field_type in self.type_spec.items()}
            result.update({'type_spec': type_spec_dict})
        return result

    @classmethod
    def _from_dict(cls, d: Dict) -> 'ColumnType':
        type_spec = None
        if 'type_spec' in d:
            type_spec = {
                field_name: cls.deserialize(field_type_dict) for field_name, field_type_dict in d['type_spec'].items()
            }
        return cls(type_spec)

    def to_sql(self) -> str:
        return 'JSONB'

    def to_sa_type(self) -> str:
        return sql.dialects.postgresql.JSONB

    def to_tf(self) -> Union['tf.TypeSpec', Dict[str, 'tf.TypeSpec']]:
        if self.type_spec is None:
            raise TypeError(f'Cannot convert {self.__class__.__name__} with missing type spec to TensorFlow')
        return {k: v.to_tf() for k, v in self.type_spec.items()}


class ArrayType(ColumnType):
    def __init__(
            self, shape: Tuple[Union[int, None], ...], dtype: ColumnType.Type):
        super().__init__(self.Type.ARRAY)
        self.shape = shape
        self.dtype = dtype

    def _supertype(cls, type1: 'ArrayType', type2: 'ArrayType') -> Optional['ArrayType']:
        if len(type1.shape) != len(type2.shape):
            return None
        base_type = ColumnType.supertype(type1.dtype, type2.dtype)
        if base_type is None:
            return None
        shape = [n1 if n1 == n2 else None for n1, n2 in zip(type1.shape, type2.shape)]
        return ArrayType(tuple(shape), base_type)

    def _as_dict(self) -> Dict:
        result = super()._as_dict()
        result.update(shape=list(self.shape), dtype=self.dtype.value)
        return result

    def __str__(self) -> str:
        return f'{self.__class__.__name__}({self.shape}, dtype={self.dtype.name})'

    @classmethod
    def _from_dict(cls, d: Dict) -> 'ColumnType':
        assert 'shape' in d
        assert 'dtype' in d
        shape = tuple(d['shape'])
        dtype = cls.Type(d['dtype'])
        return cls(shape, dtype)

    def to_sql(self) -> str:
        return 'BYTEA'

    def to_sa_type(self) -> str:
        return sql.VARBINARY

    def to_tf(self) -> Union['tf.TypeSpec', Dict[str, 'tf.TypeSpec']]:
        import tensorflow as tf
        return tf.TensorSpec(shape=self.shape, dtype=self.dtype.to_tf())


class ImageType(ColumnType):
    @enum.unique
    class Mode(enum.Enum):
        L = 0,
        RGB = 1

        @classmethod
        def from_pil(cls, pil_mode: str) -> 'Mode':
            if pil_mode == 'L':
                return cls.L
            if pil_mode == 'RGB':
                return cls.RGB

        def to_pil(self) -> str:
            return self.name

        def num_channels(self) -> int:
            return len(self.name)

    def __init__(
            self, width: Optional[int] = None, height: Optional[int] = None, size: Optional[Tuple[int, int]] = None,
            mode: Optional[Mode] = None
    ):
        """
        TODO: does it make sense to specify only width or height?
        """
        super().__init__(self.Type.IMAGE)
        assert not(width is not None and size is not None)
        assert not(height is not None and size is not None)
        if size is not None:
            self.width = size[0]
            self.height = size[1]
        else:
            self.width = width
            self.height = height
        self.mode = mode

    @property
    def num_channels(self) -> Optional[int]:
        return None if self.mode is None else self.mode.num_channels()

    def _as_dict(self) -> Dict:
        result = super()._as_dict()
        result.update(width=self.width, height=self.height, mode=self.mode.value if self.mode is not None else None)
        return result

    @classmethod
    def _from_dict(cls, d: Dict) -> 'ColumnType':
        assert 'width' in d
        assert 'height' in d
        assert 'mode' in d
        mode_val = d['mode']
        return cls(width=d['width'], height=d['height'], mode=cls.Mode(mode_val) if mode_val is not None else None)

    def conversion_fn(self, target: ColumnType) -> Optional[Callable[[Any], Any]]:
        if not target.is_image_type():
            return None
        assert isinstance(target, ImageType)
        if (target.width is None) != (target.height is None):
            # we can't resize only one dimension
            return None
        if (target.width == self.width or target.width is None) \
            and (target.height == self.height or target.height is None) \
            and (target.mode == self.mode or target.mode is None):
            # nothing to do
            return self.no_conversion
        def convert(img: PIL.Image.Image) -> PIL.Image.Image:
            if self.width != target.width or self.height != target.height:
                img = img.resize((target.width, target.height))
            if self.mode != target.mode:
                img = img.convert(target.mode.to_pil())
            return img
        return convert

    def to_sql(self) -> str:
        return 'VARCHAR'

    def to_sa_type(self) -> str:
        return sql.String

    def to_tf(self) -> Union['tf.TypeSpec', Dict[str, 'tf.TypeSpec']]:
        import tensorflow as tf
        return tf.TensorSpec(shape=(self.height, self.width, self.num_channels), dtype=tf.uint8)


class VideoType(ColumnType):
    def __init__(self):
        super().__init__(self.Type.VIDEO)

    def _as_dict(self) -> Dict:
        result = super()._as_dict()
        return result

    @classmethod
    def _from_dict(cls, d: Dict) -> 'ColumnType':
        return cls()

    def to_sql(self) -> str:
        # stored as a file path
        return 'VARCHAR'

    def to_sa_type(self) -> str:
        return sql.String

    def to_tf(self) -> Union['tf.TypeSpec', Dict[str, 'tf.TypeSpec']]:
        assert False



import sys
from typing import Optional, Callable, Dict, List, Any
import importlib
import sqlalchemy as sql
import cloudpickle

from pixeltable.type_system import ColumnType
from pixeltable import store
from pixeltable.env import Env
from pixeltable import exceptions as exc
#from pixeltable import function_pickle


class Function:
    """
    A Function's executable function is specified either directly or as module/symbol.
    In the former case, the function needs to be pickled and stored for serialization.
    In the latter case, the executable function is resolved in init().
    self.id is only set for non-module functions that are in the backing store.
    """
    def __init__(
            self, return_type: ColumnType, param_types: Optional[List[ColumnType]], id: Optional[int] = None,
            module_name: Optional[str] = None, eval_symbol: Optional[str] = None, init_symbol: Optional[str] = None,
            update_symbol: Optional[str] = None, value_symbol: Optional[str] = None,
            eval_fn: Optional[Callable] = None, init_fn: Optional[Callable] = None,
            update_fn: Optional[Callable] = None, value_fn: Optional[Callable] = None,
            order_by: List[int] = []
    ):
        has_agg_symbols = init_symbol is not None and update_symbol is not None and value_symbol is not None
        has_agg_fns = init_fn is not None and update_fn is not None and value_fn is not None
        assert (module_name is not None) == (eval_symbol is not None or has_agg_symbols)
        assert (module_name is None) == (eval_fn is not None or has_agg_fns)
        # exactly one of the 4 scenarios (is agg fn x is library fn) is specified
        assert int(eval_symbol is not None) + int(eval_fn is not None) + int(has_agg_symbols) + int(has_agg_fns) == 1
        self.return_type = return_type
        self.param_types = param_types
        self.id = id
        self.module_name = module_name
        self.eval_symbol = eval_symbol
        self.eval_fn = eval_fn
        self.init_symbol = init_symbol
        self.init_fn = init_fn
        self.update_symbol = update_symbol
        self.update_fn = update_fn
        self.value_symbol = value_symbol
        self.value_fn = value_fn

        if module_name is not None:
            # resolve module_name and symbol
            module = importlib.import_module(module_name)
            if eval_symbol is not None:
                self.eval_fn = self._resolve_symbol(module, eval_symbol)
            if init_symbol is not None:
                self.init_fn = self._resolve_symbol(module, init_symbol)
            if update_symbol is not None:
                self.update_fn = self._resolve_symbol(module, update_symbol)
            if value_symbol is not None:
                self.value_fn = self._resolve_symbol(module, value_symbol)

        if len(order_by) > 0:
            if self.init_fn is None:
                raise exc.Error(f'order_by parameter only valid for aggregate functions')
            for idx in order_by:
                if not isinstance(idx, int) or idx >= len(param_types):
                    raise exc.Error(f'order_by element not a valid index into param_types: {idx}')
        self.order_by = order_by

    def _resolve_symbol(self, module: Any, symbol: str) -> object:
        obj = module
        for el in symbol.split('.'):
            obj = getattr(obj, el)
        return obj

    @property
    def is_aggregate(self) -> bool:
        return self.init_fn is not None

    @property
    def is_library_function(self) -> bool:
        return self.module_name is not None

    def __call__(self, *args: object) -> 'pixeltable.exprs.FunctionCall':
        from pixeltable import exprs
        return exprs.FunctionCall(self, args)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self.return_type == other.return_type and self.param_types == other.param_types \
            and self.id == other.id and self.module_name == other.module_name \
            and self.eval_symbol == other.eval_symbol and self.init_symbol == other.init_symbol \
            and self.update_symbol == other.update_symbol and self.value_symbol == other.value_symbol \
            and self.eval_fn == other.eval_fn and self.init_fn == other.init_fn \
            and self.update_fn == other.update_fn and self.value_fn == other.value_fn


    def as_dict(self) -> Dict:
        if self.module_name is None and self.id is None:
            # this is not a library function and the absence of an assigned id indicates that it's not in the store yet
            FunctionRegistry.get().create_function(self)
            assert self.id is not None
        return {
            'return_type': self.return_type.as_dict(),
            'param_types': [t.as_dict() for t in self.param_types] if self.param_types is not None else None,
            'id': self.id,
            'module_name': self.module_name,
            'eval_symbol': self.eval_symbol,
            'init_symbol': self.init_symbol,
            'update_symbol': self.update_symbol,
            'value_symbol': self.value_symbol,
        }

    @classmethod
    def from_dict(cls, d: Dict) -> 'Function':
        assert 'return_type' in d
        return_type = ColumnType.from_dict(d['return_type'])
        assert 'param_types' in d
        if d['param_types'] is None:
            param_types = None
        else:
            param_types = [ColumnType.from_dict(type_dict) for type_dict in d['param_types']]
        assert 'id' in d
        assert 'module_name' in d
        assert 'eval_symbol' in d and 'init_symbol' in d and 'update_symbol' in d and 'value_symbol' in d

        if d['id'] is not None:
            assert d['module_name'] is None
            return FunctionRegistry.get().get_function(d['id'])
        else:
            return cls(
                return_type, param_types, module_name=d['module_name'], eval_symbol=d['eval_symbol'],
                init_symbol=d['init_symbol'], update_symbol=d['update_symbol'], value_symbol=d['value_symbol'])


class FunctionRegistry:
    """
    A central registry for all Functions. Handles interactions with the backing store.
    Function are loaded from the store on demand.
    """
    _instance: Optional['FunctionRegistry'] = None

    @classmethod
    def get(cls) -> 'FunctionRegistry':
        if cls._instance is None:
            cls._instance = FunctionRegistry()
        return cls._instance

    def __init__(self):
        self.fns_by_id: Dict[int, Function] = {}

    def clear_cache(self) -> None:
        """
        Useful during testing
        """
        self.fns_by_id: Dict[int, Function] = {}

    def get_function(self, id: int) -> Function:
        if id not in self.fns_by_id:
            stmt = sql.select(
                store.Function.name, store.Function.return_type, store.Function.param_types,
                store.Function.eval_obj, store.Function.init_obj, store.Function.update_obj, store.Function.value_obj) \
                .where(store.Function.id == id)
            with Env.get().engine.begin() as conn:
                rows = conn.execute(stmt)
                row = next(rows)
                name = row[0]
                return_type = ColumnType.deserialize(row[1])
                param_types = ColumnType.deserialize_list(row[2])
                eval_fn = cloudpickle.loads(row[3]) if row[3] is not None else None
                # TODO: are these checks needed?
                if row[3] is not None and eval_fn is None:
                    raise exc.Error(f'Could not load eval_fn for function {name}')
                init_fn = cloudpickle.loads(row[4]) if row[4] is not None else None
                if row[4] is not None and init_fn is None:
                    raise exc.Error(f'Could not load init_fn for aggregate function {name}')
                update_fn = cloudpickle.loads(row[5]) if row[5] is not None else None
                if row[5] is not None and update_fn is None:
                    raise exc.Error(f'Could not load update_fn for aggregate function {name}')
                value_fn = cloudpickle.loads(row[6]) if row[6] is not None else None
                if row[6] is not None and value_fn is None:
                    raise exc.Error(f'Could not load value_fn for aggregate function {name}')

                func = Function(
                    return_type, param_types, id=id, eval_fn=eval_fn, init_fn=init_fn, update_fn=update_fn,
                    value_fn=value_fn)
                self.fns_by_id[id] = func
        assert id in self.fns_by_id
        return self.fns_by_id[id]

    def create_function(
            self, fn: Function, db_id: Optional[int] = None, dir_id: Optional[int] = None,
            name: Optional[str] = None
    ) -> None:
        with Env.get().engine.begin() as conn:
            eval_fn_str = cloudpickle.dumps(fn.eval_fn) if fn.eval_fn is not None else None
            init_fn_str = cloudpickle.dumps(fn.init_fn) if fn.init_fn is not None else None
            update_fn_str = cloudpickle.dumps(fn.update_fn) if fn.update_fn is not None else None
            value_fn_str = cloudpickle.dumps(fn.value_fn) if fn.value_fn is not None else None
            res = conn.execute(
                sql.insert(store.Function.__table__)
                    .values(
                        db_id=db_id, dir_id=dir_id, name=name, return_type=fn.return_type.serialize(),
                        param_types=ColumnType.serialize_list(fn.param_types),
                        eval_obj=eval_fn_str, init_obj=init_fn_str, update_obj=update_fn_str, value_obj=value_fn_str))
            fn.id = res.inserted_primary_key[0]
            self.fns_by_id[fn.id] = fn

    def update_function(
            self, id: int, eval_fn: Optional[Callable] = None, init_fn: Optional[Callable] = None,
            update_fn: Optional[Callable] = None, value_fn: Optional[Callable] = None
    ) -> None:
        """
        Updates the callable for the function with the given id in the store and in the cache, if present.
        """
        with Env.get().engine.begin() as conn:
            updates = {}
            if eval_fn is not None:
                updates[store.Function.eval_obj] = cloudpickle.dumps(eval_fn)
            if init_fn is not None:
                updates[store.Function.init_obj] = cloudpickle.dumps(init_fn)
            if update_fn is not None:
                updates[store.Function.update_obj] = cloudpickle.dumps(update_fn)
            if value_fn is not None:
                updates[store.Function.value_obj] = cloudpickle.dumps(value_fn)
            conn.execute(
                sql.update(store.Function.__table__)
                    .values(updates)
                    .where(store.Function.id == id))
        if id in self.fns_by_id:
            if eval_fn is not None:
                self.fns_by_id[id].eval_fn = eval_fn
            if init_fn is not None:
                self.fns_by_id[id].init_fn = init_fn
            if update_fn is not None:
                self.fns_by_id[id].update_fn = update_fn
            if value_fn is not None:
                self.fns_by_id[id].value_fn = value_fn

    def delete_function(self, id: int) -> None:
        assert id is not None
        with Env.get().engine.begin() as conn:
            conn.execute(
                sql.delete(store.Function.__table__)
                    .where(store.Function.id == id))


# def create_module_list() -> None:
#     """
#     Generate file standard_modules.py, which contains a list of modules available after 'import pixeltable'.
#     These are the modules we don't want to pickle.
#     TODO: move this elsewhere?
#     """
#     with open('standard_modules.py', 'w') as f:
#         f.write('module_names = set([\n    ')
#         line_len = 0
#         module_names = sys.modules.keys()
#         for name in module_names:
#             str = f"'{name}', "
#             line_len += len(str)
#             if line_len >= 80:
#                 f.write('\n    ')
#                 line_len = 4  # spaces
#             f.write(str)
#         f.write('\n])')


# make create_module_list() callable from the commandline
if __name__ == '__main__':
    globals()[sys.argv[1]]()



import pytest
import pandas as pd

import pixeltable as pt
from pixeltable.type_system import VideoType, IntType, ImageType
from pixeltable.tests.utils import get_video_files
from pixeltable import catalog
from pixeltable import utils
from pixeltable import exceptions as exc


class TestVideo:
    def test_basic(self, test_db: catalog.Db) -> None:
        video_filepaths = get_video_files()
        db = test_db
        cols = [
            catalog.Column('video', VideoType(), nullable=False),
            catalog.Column('frame', ImageType(), nullable=False),
            catalog.Column('frame_idx', IntType(), nullable=False),
        ]
        tbl = db.create_table(
            'test', cols, extract_frames_from = 'video', extracted_frame_col = 'frame',
            extracted_frame_idx_col = 'frame_idx', extracted_fps = 1)
        tbl.insert_rows([[p] for p in video_filepaths], columns=['video'])
        assert utils.extracted_frame_count(tbl_id=tbl.id) == tbl.count()
        _ = tbl[tbl.frame_idx, tbl.frame, tbl.frame.rotate(90)].show(0)

        # missing 'columns' arg
        with pytest.raises(exc.Error):
            tbl.insert_rows([[p] for p in video_filepaths])

        # column values mismatch in rows
        with pytest.raises(exc.Error):
            tbl.insert_rows([[1, 2], [3]], columns=['video'])

        # column values mismatch in rows
        with pytest.raises(exc.Error):
            tbl.insert_rows([[1, 2]], columns=['video'])

        # revert() also removes extracted frames
        tbl.revert()
        assert utils.extracted_frame_count(tbl_id=tbl.id) == tbl.count()
